from __future__ import annotations

import zospy as zp


def _initialize_system(
    oss: zp.zpcore.OpticStudioSystem,
    aperture_type: zp.constants.SystemData.ZemaxApertureType | str = "FloatByStopSize",
    wavelength: float = 543,
):
    oss.new()

    oss.SystemData.Aperture.ApertureType = zp.constants.process_constant(
        zp.constants.SystemData.ZemaxApertureType, aperture_type
    )
    oss.SystemData.Wavelengths.GetWavelength(1).Wavelength = wavelength / 1000


def empty_system(oss: zp.zpcore.OpticStudioSystem) -> zp.zpcore.OpticStudioSystem:
    _initialize_system(oss)

    return oss


def simple_system(oss: zp.zpcore.OpticStudioSystem) -> zp.zpcore.OpticStudioSystem:
    _initialize_system(oss)

    surface_object = oss.LDE.GetSurfaceAt(0)
    surface_object.Thickness = float("inf")

    surface_stop = oss.LDE.GetSurfaceAt(1)
    surface_stop.SemiDiameter = 1

    lens_front = oss.LDE.InsertNewSurfaceAt(2)
    lens_front.Comment = "lens front"
    lens_front.Radius = 20
    lens_front.Thickness = 1
    zp.solvers.material_model(lens_front.MaterialCell, refractive_index=1.5)

    lens_back = oss.LDE.InsertNewSurfaceAt(3)
    lens_back.Comment = "lens back"
    lens_back.Radius = -20
    lens_back.Thickness = 19.792

    return oss


def polarized_system(oss: zp.zpcore.OpticStudioSystem) -> zp.zpcore.OpticStudioSystem:
    oss = simple_system(oss)

    lens_front = oss.LDE.GetSurfaceAt(2)

    # Change surface to Jones Matrix
    zp.functions.lde.surface_change_type(lens_front, zp.constants.Editors.LDE.SurfaceType.JonesMatrix)

    # Set Jones Matrix elements to horizontal polarization
    lens_front.GetCellAt(12).DoubleValue = 1  # A real
    lens_front.GetCellAt(18).DoubleValue = 0  # D real

    return oss


def decentered_system(oss: zp.zpcore.OpticStudioSystem) -> zp.zpcore.OpticStudioSystem:
    oss = simple_system(oss)

    field = oss.SystemData.Fields.GetField(1)
    field.X = 10
    field.Y = 20

    return oss



import re
from inspect import getsource

import pytest

from zospy.analyses.base import BaseAnalysisWrapper, new_analysis
from zospy.analyses.systemviewers.base import SystemViewerWrapper
from zospy.api import constants
from zospy.api.constants import process_constant

analysis_wrapper_classes = BaseAnalysisWrapper.__subclasses__()
analysis_wrapper_classes.remove(SystemViewerWrapper)

REGEX_SETTING = re.compile(r"\s*self\.analysis\.Settings\.(?P<setting>\w+)")


@pytest.mark.parametrize("analysis_wrapper", analysis_wrapper_classes)
def test_settings_exist(empty_system, analysis_wrapper):
    if analysis_wrapper.MODE == "Nonsequential":
        empty_system.make_nonsequential()

    analysis = new_analysis(empty_system, process_constant(constants.Analysis.AnalysisIDM, analysis_wrapper.TYPE))
    source = getsource(analysis_wrapper.run_analysis)

    settings = REGEX_SETTING.findall(source)

    nonexistent_settings = [setting for setting in settings if not hasattr(analysis.Settings, setting)]

    assert not nonexistent_settings, f"Nonexistent settings: {nonexistent_settings}"



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("XYSymmetricField", "YSymmetricField")

class XYSymmetricField:
    All: XYSymmetricField = None
    Zero: XYSymmetricField = None
    YPositive70Percent: XYSymmetricField = None
    YNegative70Percent: XYSymmetricField = None
    YPositive100Percent: XYSymmetricField = None
    YNegative100Percent: XYSymmetricField = None
    XPositive70Percent: XYSymmetricField = None
    XNegative70Percent: XYSymmetricField = None
    XPositive100Percent: XYSymmetricField = None
    XNegative100Percent: XYSymmetricField = None

class YSymmetricField:
    All: YSymmetricField = None
    Zero: YSymmetricField = None
    Positive70Percent: YSymmetricField = None
    Negative70Percent: YSymmetricField = None
    Positive100Percent: YSymmetricField = None
    Negative100Percent: YSymmetricField = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "DisplayAsTypes",
    "Distortions",
    "FFA_AberrationTypes",
    "FFA_DecompositionTypes",
    "FFA_DisplayTypes",
    "FFA_FieldShapes",
    "FFA_ShowAsTypes",
    "FieldScanDirections",
    "RayTraceType",
    "ZernikeCoefficientTypes",
)

class DisplayAsTypes:
    Percent: DisplayAsTypes = None
    Absolute: DisplayAsTypes = None

class Distortions:
    F_TanTheta: Distortions = None
    F_Theta: Distortions = None
    Cal_F_Theta: Distortions = None
    Cal_F_TanTheta: Distortions = None
    SMIA_TV: Distortions = None

class FFA_AberrationTypes:
    Defocus: FFA_AberrationTypes = None
    PrimaryAstigmatism: FFA_AberrationTypes = None
    PrimaryComa: FFA_AberrationTypes = None
    SpecifiedTerm: FFA_AberrationTypes = None

class FFA_DecompositionTypes:
    ZernikeTerms: FFA_DecompositionTypes = None

class FFA_DisplayTypes:
    Absolute: FFA_DisplayTypes = None
    Relative: FFA_DisplayTypes = None
    Average: FFA_DisplayTypes = None

class FFA_FieldShapes:
    Rectangular: FFA_FieldShapes = None
    Elliptical: FFA_FieldShapes = None

class FFA_ShowAsTypes:
    GreyScale: FFA_ShowAsTypes = None
    GreyScaleInverted: FFA_ShowAsTypes = None
    FalseColor: FFA_ShowAsTypes = None
    FalseColorInverted: FFA_ShowAsTypes = None
    Icons: FFA_ShowAsTypes = None

class FieldScanDirections:
    Plus_Y: FieldScanDirections = None
    Plus_X: FieldScanDirections = None
    Minus_Y: FieldScanDirections = None
    Minus_X: FieldScanDirections = None

class RayTraceType:
    DirectionCosines: RayTraceType = None
    TangentAngle: RayTraceType = None
    YmUmYcUc: RayTraceType = None

class ZernikeCoefficientTypes:
    Fringe: ZernikeCoefficientTypes = None
    Standard: ZernikeCoefficientTypes = None
    Annular: ZernikeCoefficientTypes = None



"""Huygens MTF analysis."""

from __future__ import annotations

from typing import Annotated, Union

from pandas import DataFrame
from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_settings
from zospy.analyses.parsers.types import FieldNumber, WavelengthNumber, ZOSAPIConstant  # noqa: TCH001
from zospy.api import constants
from zospy.utils.zputils import standardize_sampling

__all__ = ("HuygensMTF", "HuygensMtfSettings")


@analysis_settings
class HuygensMtfSettings:
    """Settings for the Huygens MTF analysis.

    For an in depth explanation of the parameters, see the OpticStudio user manual.

    Attributes
    ----------
    pupil_sampling : str | int
        The pupil sampling, either string (e.g. '64x64') or int.
        The integer will be treated as a ZOSAPI Constants integer.
    image_sampling : str | int
        The image sampling, either string (e.g., '64x64') or int.
        The integer will be treated as a ZOSAPI Constants integer.
    image_delta : float
        The Image Delta, defaults to 0.0.
    wavelength : str | int
        The wavelength to use in the MTF. Either 'All' or an integer specifying the wavelength number.
    field : str | int
        The field to use in the MTF. Either 'All' or an integer specifying the field number.
    mtf_type : ZOSAPI.Analysis.Settings.Mtf.MtfTypes.Modulation
        The MTF type (e.g. `Modulation`) that is calculated.
    maximum_frequency : float
        The maximum frequency at which the MTF is calculated.
        Units are either cycles/mm or cycles/mr, depending on system setting.
        Defaults to 150.0, which is more appropriate when units are set to cycles/mm.
    use_polarization : bool
        Use polarization. Defaults to False.
    use_dashes : bool
        Use dashes. Defaults to False.
    """

    pupil_sampling: str | Annotated[int, Field(ge=0)] = Field(default="32x32", description="Pupil sampling grid size")
    image_sampling: str | Annotated[int, Field(ge=0)] = Field(default="32x32", description="Image sampling grid size")
    image_delta: float = Field(default=0.0, description="Image delta")
    wavelength: WavelengthNumber = Field(default="All", description="Wavelength number or 'All'")
    field: FieldNumber = Field(default="All", description="Field number or 'All'")
    mtf_type: ZOSAPIConstant("Analysis.Settings.Mtf.HuygensMtfTypes") = Field(
        default="Modulation", description="MTF type"
    )
    maximum_frequency: float = Field(default=150.0, description="Maximum frequency")
    use_polarization: bool = Field(default=False, description="Use polarization")
    use_dashes: bool = Field(default=False, description="Use dashes")


class HuygensMTF(BaseAnalysisWrapper[Union[DataFrame, None], HuygensMtfSettings], analysis_type="HuygensMtf"):
    """Huygens Modulation Transfer Function (MTF) analysis."""

    def __init__(
        self,
        *,
        pupil_sampling: str | int = "32x32",
        image_sampling: str | int = "32x32",
        image_delta: float = 0.0,
        wavelength: int | str = "All",
        field: int | str = "All",
        mtf_type: str | constants.Analysis.Settings.Mtf.HuygensMtfTypes = "Modulation",
        maximum_frequency: float = 150.0,
        use_polarization: bool = False,
        use_dashes: bool = False,
    ):
        """Create a new Huygens MTF analysis.

        See Also
        --------
        HuygensMtfSettings : Settings for the Huygens MTF analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> DataFrame | None:
        """Run the Huygens MTF analysis."""
        self.analysis.Settings.PupilSampleSize = getattr(
            constants.Analysis.SampleSizes, standardize_sampling(self.settings.pupil_sampling)
        )
        self.analysis.Settings.ImageSampleSize = getattr(
            constants.Analysis.SampleSizes, standardize_sampling(self.settings.image_sampling)
        )
        self.analysis.Settings.ImageDelta = self.settings.image_delta
        self.analysis.set_wavelength(self.settings.wavelength)
        self.analysis.set_field(self.settings.field)
        self.analysis.Settings.Type = constants.process_constant(
            constants.Analysis.Settings.Mtf.HuygensMtfTypes, self.settings.mtf_type
        )
        self.analysis.Settings.MaximumFrequency = self.settings.maximum_frequency
        self.analysis.Settings.UsePolarization = self.settings.use_polarization
        self.analysis.Settings.UseDashes = self.settings.use_dashes

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Get results
        return self.get_data_series()



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import Beam, IMessage, IMessages, SampleSizes, ShowAs

from . import (
    RMS,
    Aberrations,
    Coatings,
    DiffractionEfficiency,
    EncircledEnergy,
    ExtendedScene,
    Fans,
    Materials,
    Mtf,
    NSCSurface,
    Psf,
    RayTracing,
    Spot,
    Surface,
    Wavefront,
)

__all__ = (
    "Aberrations",
    "Coatings",
    "DiffractionEfficiency",
    "EncircledEnergy",
    "ExtendedScene",
    "Fans",
    "Materials",
    "Mtf",
    "NSCSurface",
    "Psf",
    "RayTracing",
    "RMS",
    "Spot",
    "Surface",
    "Wavefront",
    "AxisType",
    "DetectorViewerScaleTypes",
    "DisplayOption",
    "HuygensPsfTypes",
    "IAS_",
    "IAS_CriticalRayTracer",
    "IAS_Detector",
    "IAS_Field",
    "IAS_FileComparator",
    "IAS_Interferogram",
    "IAS_Surface",
    "IAS_WavefrontMap",
    "IAS_Wavelength",
    "IAS_XXXTemplateXXX",
    "Parity",
    "Polarizations",
    "PsfSpread",
    "PsfTypes",
    "ReferenceGia",
    "Rotations",
    "ScanTypes",
    "SourceGia",
    "STAREffectsOptions",
)

class AxisType:
    X = 0
    Y = 1
    Z = 2

class DetectorViewerScaleTypes:
    Linear = 0
    Log_Minus_5 = 1
    Normalized = 1
    Log_Minus_10 = 2
    Log_Minus_15 = 3

class DisplayOption:
    AllRays = 0
    FailedRays = 1
    PassedRays = 2

class HuygensPsfTypes:
    Linear = 0
    Log_Minus_1 = 1
    Log_Minus_2 = 2
    Log_Minus_3 = 3
    Log_Minus_4 = 4
    Log_Minus_5 = 5
    Real = 6
    Imaginary = 7
    Phase = 8

class IAS_:
    def Load(self) -> None: ...
    def LoadFrom(self, settingsFile: str) -> bool: ...
    def ModifySettings(self, settingsFile: str, typeCode: str, newValue: str) -> bool: ...
    def Reset(self) -> None: ...
    def Save(self) -> None: ...
    def SaveTo(self, settingsFile: str) -> bool: ...
    def Verify(self) -> IMessages: ...

class IAS_CriticalRayTracer(IAS_):
    @property
    def PositionTolerance(self) -> float: ...
    @PositionTolerance.setter
    def PositionTolerance(self, value: float) -> None: ...
    @property
    def AngleTolerance(self) -> float: ...
    @AngleTolerance.setter
    def AngleTolerance(self, value: float) -> None: ...
    @property
    def Filename(self) -> str: ...
    @Filename.setter
    def Filename(self, value: str) -> None: ...
    @property
    def RaysToDisplay(self) -> DisplayOption: ...
    @RaysToDisplay.setter
    def RaysToDisplay(self, value: DisplayOption) -> None: ...
    @property
    def DisplayXYZ(self) -> bool: ...
    @DisplayXYZ.setter
    def DisplayXYZ(self, value: bool) -> None: ...
    @property
    def DisplayLMN(self) -> bool: ...
    @DisplayLMN.setter
    def DisplayLMN(self, value: bool) -> None: ...
    @property
    def DisplayActualEndData(self) -> bool: ...
    @DisplayActualEndData.setter
    def DisplayActualEndData(self, value: bool) -> None: ...
    @property
    def DisplayTargetEndData(self) -> bool: ...
    @DisplayTargetEndData.setter
    def DisplayTargetEndData(self, value: bool) -> None: ...
    @property
    def DisplayStartData(self) -> bool: ...
    @DisplayStartData.setter
    def DisplayStartData(self, value: bool) -> None: ...

class IAS_Detector:
    def GetDetectorNumber(self) -> int: ...
    def SetDetectorNumber(self, N: int) -> IMessage: ...
    def SummarizeAll(self) -> IMessage: ...

class IAS_Field:
    def GetFieldNumber(self) -> int: ...
    def SetFieldNumber(self, N: int) -> IMessage: ...
    def UseAllFields(self) -> IMessage: ...

class IAS_FileComparator(IAS_):
    @property
    def File1Name(self) -> str: ...
    @File1Name.setter
    def File1Name(self, value: str) -> None: ...
    @property
    def File1SpecifyConfig(self) -> bool: ...
    @File1SpecifyConfig.setter
    def File1SpecifyConfig(self, value: bool) -> None: ...
    @property
    def File1Config(self) -> int: ...
    @File1Config.setter
    def File1Config(self, value: int) -> None: ...
    @property
    def File2Name(self) -> str: ...
    @File2Name.setter
    def File2Name(self, value: str) -> None: ...
    @property
    def File2SpecifyConfig(self) -> bool: ...
    @File2SpecifyConfig.setter
    def File2SpecifyConfig(self, value: bool) -> None: ...
    @property
    def File2Config(self) -> int: ...
    @File2Config.setter
    def File2Config(self, value: int) -> None: ...
    @property
    def IncludeGeneralData(self) -> bool: ...
    @IncludeGeneralData.setter
    def IncludeGeneralData(self, value: bool) -> None: ...
    @property
    def IncludeSurfaceData(self) -> bool: ...
    @IncludeSurfaceData.setter
    def IncludeSurfaceData(self, value: bool) -> None: ...
    @property
    def IncludeSurfaceDetail(self) -> bool: ...
    @IncludeSurfaceDetail.setter
    def IncludeSurfaceDetail(self, value: bool) -> None: ...
    @property
    def IncludeEdgeThickness(self) -> bool: ...
    @IncludeEdgeThickness.setter
    def IncludeEdgeThickness(self, value: bool) -> None: ...
    @property
    def IncludeMultiConfigData(self) -> bool: ...
    @IncludeMultiConfigData.setter
    def IncludeMultiConfigData(self, value: bool) -> None: ...
    @property
    def IncludeSolvesVariables(self) -> bool: ...
    @IncludeSolvesVariables.setter
    def IncludeSolvesVariables(self, value: bool) -> None: ...
    @property
    def IncludeIndexTCEData(self) -> bool: ...
    @IncludeIndexTCEData.setter
    def IncludeIndexTCEData(self, value: bool) -> None: ...
    @property
    def IncludeGlobalVertex(self) -> bool: ...
    @IncludeGlobalVertex.setter
    def IncludeGlobalVertex(self, value: bool) -> None: ...
    @property
    def IncludeCOCPoint(self) -> bool: ...
    @IncludeCOCPoint.setter
    def IncludeCOCPoint(self, value: bool) -> None: ...
    @property
    def IncludeElementVolume(self) -> bool: ...
    @IncludeElementVolume.setter
    def IncludeElementVolume(self, value: bool) -> None: ...
    @property
    def IncludeFNumbers(self) -> bool: ...
    @IncludeFNumbers.setter
    def IncludeFNumbers(self, value: bool) -> None: ...
    @property
    def IncludeCardinalPoints(self) -> bool: ...
    @IncludeCardinalPoints.setter
    def IncludeCardinalPoints(self, value: bool) -> None: ...
    @property
    def IncludePOPSettings(self) -> bool: ...
    @IncludePOPSettings.setter
    def IncludePOPSettings(self, value: bool) -> None: ...
    @property
    def IncludeFilesUsed(self) -> bool: ...
    @IncludeFilesUsed.setter
    def IncludeFilesUsed(self, value: bool) -> None: ...
    @property
    def IncludeMeritFunction(self) -> bool: ...
    @IncludeMeritFunction.setter
    def IncludeMeritFunction(self, value: bool) -> None: ...
    @property
    def ShowDifferencesOnly(self) -> bool: ...
    @ShowDifferencesOnly.setter
    def ShowDifferencesOnly(self, value: bool) -> None: ...
    def ClearAllIncludes(self) -> None: ...
    def SetAllIncludes(self) -> None: ...

class IAS_Interferogram(IAS_):
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def Sampling(self) -> SampleSizes: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes) -> None: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @property
    def Beam_1(self) -> Beam: ...
    @Beam_1.setter
    def Beam_1(self, value: Beam) -> None: ...
    @property
    def Beam_2(self) -> Beam: ...
    @Beam_2.setter
    def Beam_2(self, value: Beam) -> None: ...
    @property
    def Ref_Beam_1_To_Vertex(self) -> bool: ...
    @Ref_Beam_1_To_Vertex.setter
    def Ref_Beam_1_To_Vertex(self, value: bool) -> None: ...
    @property
    def Ref_Beam_2_To_Vertex(self) -> bool: ...
    @Ref_Beam_2_To_Vertex.setter
    def Ref_Beam_2_To_Vertex(self, value: bool) -> None: ...
    @property
    def UseExitPupil(self) -> bool: ...
    @UseExitPupil.setter
    def UseExitPupil(self, value: bool) -> None: ...
    @property
    def ConsiderOPL(self) -> bool: ...
    @ConsiderOPL.setter
    def ConsiderOPL(self, value: bool) -> None: ...
    @property
    def ScaleFactor(self) -> float: ...
    @ScaleFactor.setter
    def ScaleFactor(self, value: float) -> None: ...
    @property
    def X_Tilt(self) -> float: ...
    @X_Tilt.setter
    def X_Tilt(self, value: float) -> None: ...
    @property
    def Y_Tilt(self) -> float: ...
    @Y_Tilt.setter
    def Y_Tilt(self, value: float) -> None: ...
    @property
    def Subaperture_X(self) -> float: ...
    @Subaperture_X.setter
    def Subaperture_X(self, value: float) -> None: ...
    @property
    def Subaperture_Y(self) -> float: ...
    @Subaperture_Y.setter
    def Subaperture_Y(self, value: float) -> None: ...
    @property
    def Subaperture_R(self) -> float: ...
    @Subaperture_R.setter
    def Subaperture_R(self, value: float) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...

class IAS_Surface:
    def GetSurfaceNumber(self) -> int: ...
    def SetSurfaceNumber(self, N: int) -> IMessage: ...
    def UseImageSurface(self) -> IMessage: ...
    def UseObjectiveSurface(self) -> IMessage: ...

class IAS_WavefrontMap(IAS_):
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @property
    def Rotation(self) -> Rotations: ...
    @Rotation.setter
    def Rotation(self, value: Rotations) -> None: ...
    @property
    def Sampling(self) -> SampleSizes: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes) -> None: ...
    @property
    def Polarization(self) -> Polarizations: ...
    @Polarization.setter
    def Polarization(self, value: Polarizations) -> None: ...
    @property
    def ReferenceToPrimary(self) -> bool: ...
    @ReferenceToPrimary.setter
    def ReferenceToPrimary(self, value: bool) -> None: ...
    @property
    def UseExitPupil(self) -> bool: ...
    @UseExitPupil.setter
    def UseExitPupil(self, value: bool) -> None: ...
    @property
    def RemoveTilt(self) -> bool: ...
    @RemoveTilt.setter
    def RemoveTilt(self, value: bool) -> None: ...
    @property
    def Scale(self) -> float: ...
    @Scale.setter
    def Scale(self, value: float) -> None: ...
    @property
    def Subaperture_X(self) -> float: ...
    @Subaperture_X.setter
    def Subaperture_X(self, value: float) -> None: ...
    @property
    def Subaperture_Y(self) -> float: ...
    @Subaperture_Y.setter
    def Subaperture_Y(self, value: float) -> None: ...
    @property
    def Subaperture_R(self) -> float: ...
    @Subaperture_R.setter
    def Subaperture_R(self, value: float) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def STAREffects(self) -> STAREffectsOptions: ...
    @STAREffects.setter
    def STAREffects(self, value: STAREffectsOptions) -> None: ...

class IAS_Wavelength:
    def GetWavelengthNumber(self) -> int: ...
    def SetWavelengthNumber(self, N: int) -> IMessage: ...
    def UseAllWavelengths(self) -> IMessage: ...

class IAS_XXXTemplateXXX(IAS_): ...

class Parity:
    Even = 0
    Odd = 1

class Polarizations:
    # None = 0
    Ex = 1
    Ey = 2
    Ez = 3

class PsfSpread:
    Line = 0
    Edge = 1

class PsfTypes:
    X_Linear = 0
    Y_Linear = 1
    X_Logarithmic = 2
    Y_Logarithmic = 3
    X_Phase = 4
    Y_Phase = 5
    X_RealPart = 6
    Y_RealPart = 7
    X_ImaginaryPart = 8
    Y_ImaginaryPart = 9

class ReferenceGia:
    ChiefRay = 0
    Vertex = 1
    PrimaryChief = 2
    Centroid = 3

class Rotations:
    Rotate_0 = 0
    Rotate_90 = 1
    Rotate_180 = 2
    Rotate_270 = 3

class ScanTypes:
    Plus_Y = 0
    Plus_X = 1
    Minus_Y = 2
    Minus_X = 3

class SourceGia:
    Uniform = 0
    Lambertian = 1

class STAREffectsOptions:
    On = 0
    Difference = 1



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("FileSource", "IFile", "KnownFolder", "Operation")

class FileSource:
    Project = 0
    Root = 1
    Unknown = -1

class IFile:
    @property
    def Name(self) -> str: ...
    @property
    def KnownFolder(self) -> KnownFolder: ...
    @property
    def RelativePath(self) -> str: ...
    @property
    def Source(self) -> FileSource: ...

class KnownFolder:
    Project = 0
    ABgData = 1
    BlackBoxes = 2
    Coatings = 3
    Configs = 4
    DllBulkScatter = 5
    DllDiffractive = 6
    DllGradientIndex = 7
    DllObjects = 8
    DllPhysicalOptics = 9
    DllSources = 10
    DllSurfaceScatter = 11
    DllSurfaces = 12
    Extend = 13
    GlassCat = 14
    ImaFiles = 15
    Macros = 16
    MeritFunction = 17
    Miscellaneous = 18
    ObjectsApertures = 19
    ObjectsCadFiles = 20
    ObjectsCreoParametricFiles = 21
    ObjectsGridFiles = 22
    ObjectsInventorFiles = 23
    ObjectsPartDesignerObjects = 24
    ObjectsPhosphorsAndFluorescenceFiles = 25
    ObjectsPolygonObjects = 26
    ObjectsStopFiles = 27
    ObjectsSolidWorksFiles = 28
    ObjectsSourcesEulumdat = 29
    ObjectsSourcesIesna = 30
    ObjectsSourcesRadiantSourceModelFiles = 31
    ObjectsSourcesSourceFiles = 32
    ObjectsSourcesSpectrumFiles = 33
    ObjectsTabulatedObjects = 34
    PopBeamFiles = 35
    Profiles = 36
    ScatterData = 37
    Tolerance = 38
    Udo = 39
    ZosApiExtensions = 40
    ZosApiOperands = 41
    ZosApiUserAnalysis = 42
    DllBirefringence = 43
    Unknown = -1

class Operation:
    # None = 0
    GetFilesInUse = 1
    GetAdditionalFiles = 2



import pytest

from zospy.functions.lde import find_surface_by_comment


@pytest.mark.parametrize(
    "comment,case_sensitive,expected_indices",
    [
        ("Abcd", True, [0]),  # When using .SurfaceNumber, the returned value matches the LDE numbering (starting at 0)
        ("Abcd", False, [0, 1, 2, 3]),
        ("ABcd", True, [1]),
        ("ABCD", True, []),
        ("ABCD", False, [0, 1, 2, 3]),
        ("efgh", True, []),
        ("efgh", False, []),
    ],
)
def test_can_find_surface_by_comment(simple_system, comment, case_sensitive, expected_indices):
    simple_system.LDE.GetSurfaceAt(0).Comment = "Abcd"
    simple_system.LDE.GetSurfaceAt(1).Comment = "ABcd"
    simple_system.LDE.GetSurfaceAt(2).Comment = "ABCd"
    simple_system.LDE.GetSurfaceAt(3).Comment = "ABCd"

    result = find_surface_by_comment(simple_system.LDE, comment=comment, case_sensitive=case_sensitive)

    indices = [surface.SurfaceNumber for surface in result]

    assert indices == expected_indices



"""OpticStudio analyses from the Wavefront category."""

from zospy.analyses.wavefront.wavefront_map import WavefrontMap, WavefrontMapSettings
from zospy.analyses.wavefront.zernike_standard_coefficients import (
    ZernikeStandardCoefficients,
    ZernikeStandardCoefficientsSettings,
)

__all__ = ("ZernikeStandardCoefficients", "ZernikeStandardCoefficientsSettings", "WavefrontMap", "WavefrontMapSettings")



"""Nonsequential 3D Layout viewer."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field

from zospy.analyses.decorators import analysis_settings
from zospy.analyses.systemviewers.base import ImageSize, SystemViewerWrapper

if TYPE_CHECKING:
    from zospy.api import _ZOSAPI

__all__ = ("NSC3DLayout", "NSC3DLayoutSettings")


@analysis_settings
class NSC3DLayoutSettings:
    """Settings for the nonsequential 3D Layout viewer.

    Notes
    -----
    Not all settings available in OpticStudio are supported yet.
    """

    image_size: ImageSize = Field(default=(800, 600), description="Image size")


class NSC3DLayout(SystemViewerWrapper[NSC3DLayoutSettings], analysis_type="NSC3DLayout", mode="Nonsequential"):
    """3D Layout viewer for Non-Sequential systems."""

    def __init__(self, *, image_size: tuple[int, int] = (800, 600)):
        """Create a new nonsequential 3D Layout viewer.

        See Also
        --------
        NSC3DLayoutSettings : Settings for the NSC 3D Layout viewer
        """
        super().__init__(settings_kws=locals())

    def configure_layout_tool(
        self,
    ) -> _ZOSAPI.Tools.Layouts.INSC3DLayoutExport:
        """Configure the nonsequential 3D Layout viewer."""
        layout_tool = self.oss.Tools.Layouts.OpenNSC3DLayoutExport()

        layout_tool.OutputPixelWidth, layout_tool.OutputPixelHeight = self.settings.image_size

        return layout_tool



"""Single Ray Trace analysis."""

from __future__ import annotations

from typing import Literal
from warnings import warn

import pandas as pd
from pydantic import Field, model_validator

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers.transformers import ZospyTransformer
from zospy.analyses.parsers.types import UnitField, ValidatedDataFrame  # noqa: TCH001
from zospy.api import constants

__all__ = ("SingleRayTrace", "SingleRayTraceSettings")


class SingleRayTraceTransformer(ZospyTransformer):
    """Transformer for the output of the Single Ray Trace analysis."""

    def NAN(self, args):  # noqa: N802, ARG002
        """Transform a NaN value."""
        return float("nan")

    def ray_trace_data_table(self, args):
        """Transform the ray trace data table to a DataFrame."""
        header, rows = args[0]
        header_length = len(header)

        # Fill empty columns with NaN
        for row in rows:
            if (row_length := len(row)) < header_length:
                # Warning is only raised once for the full loop
                warn("Header and row length mismatch. Empty columns will be filled with NaN.")

                # Check if the last row value is a comment
                if isinstance(row[-1], str):
                    row[-1:-1] = [float("nan")] * (header_length - row_length)
                else:
                    # Insert empty columns at the end of the row
                    row.extend([float("nan")] * (header_length - row_length))

        return pd.DataFrame(rows, columns=header)


@analysis_result
class SingleRayTraceResult:
    """Data for the Single Ray Trace analysis."""

    units: str = Field(alias="Units")
    coordinates: str = Field(alias="Coordinates")
    wavelength: UnitField[float] = Field(alias="Wavelength")

    normalized_x_field_coord: float | None = Field(alias="Normalized X Field Coord (Hx)", default=None)
    normalized_y_field_coord: float | None = Field(alias="Normalized Y Field Coord (Hy)", default=None)
    normalized_x_pupil_coord: float | None = Field(alias="Normalized X Pupil Coord (Px)", default=None)
    normalized_y_pupil_coord: float | None = Field(alias="Normalized Y Pupil Coord (Py)", default=None)

    real_ray_trace_data: ValidatedDataFrame | None = Field(alias="Real Ray Trace Data", default=None)
    paraxial_ray_trace_data: ValidatedDataFrame | None = Field(alias="Paraxial Ray Trace Data", default=None)
    ym_um_yc_uc_ray_trace_data: ValidatedDataFrame | None = Field(
        alias="Trace of Paraxial Y marginal, U marginal, Y chief, U chief only.",
        default=None,
    )

    @model_validator(mode="after")
    def validate_ray_trace_data(self):
        """Check if all ray trace data types belonging to the ray trace type are present."""
        if (
            self.real_ray_trace_data is None
            and self.paraxial_ray_trace_data is None
            and self.ym_um_yc_uc_ray_trace_data is None
        ):
            raise ValueError("Ray trace data must be specified")
        if (self.real_ray_trace_data is not None and self.paraxial_ray_trace_data is None) or (
            self.real_ray_trace_data is None and self.paraxial_ray_trace_data is not None
        ):
            raise ValueError("Both real and paraxial ray trace data must be specified")
        if self.ym_um_yc_uc_ray_trace_data is not None and (
            self.real_ray_trace_data is not None or self.paraxial_ray_trace_data is not None
        ):
            raise ValueError("Either specify real and paraxial ray trace data or Ym, Um, Yc, Uc ray trace data")

        return self


@analysis_settings
class SingleRayTraceSettings:
    """Settings for the Single Ray Trace analysis.

    Attributes
    ----------
    hx : float
        Normalized X field coordinate. Defaults to 0.
    hy : float
        Normalized Y field coordinate. Defaults to 0.
    px : float
        Normalized X pupil coordinate. Defaults to 0.
    py : float
        Normalized Y pupil coordinate. Defaults to 1.
    wavelength : int
        The wavelength number that is to be used. Must be an integer specifying the wavelength number.
        Defaults to 1.
    field : int
        The field number that is to be used. Must be an integer specifying the field number. "Arbitrary" is 0. Defaults
        to 0.
    raytrace_type : zospy.api.constants.Analysis.Settings.Aberrations.RayTraceType
        The Ray Trace type (e.g. 'DirectionCosines') that is calculated. Defaults to 'DirectionCosines'.
    global_coordinates : bool
        Defines if global coordinates are used. Defaults to `False`.
    """

    # TODO: document the fields
    hx: float = Field(ge=-1, le=1, default=0, description="Normalized X field coordinate")
    hy: float = Field(ge=-1, le=1, default=0, description="Normalized Y field coordinate")
    px: float = Field(ge=-1, le=1, default=0, description="Normalized X pupil coordinate")
    py: float = Field(ge=-1, le=1, default=0, description="Normalized Y pupil coordinate")
    wavelength: int = Field(gt=0, default=1, description="Wavelength number")
    field: int = Field(ge=0, default=0, description="Field number")
    raytrace_type: Literal["DirectionCosines", "TangentAngle", "YmUmYcUc"] = Field(default="DirectionCosines")
    global_coordinates: bool = Field(default=False, description="Use global coordinates")


class SingleRayTrace(
    BaseAnalysisWrapper[SingleRayTraceResult, SingleRayTraceSettings],
    analysis_type="RayTrace",
    needs_text_output_file=True,
):
    """Single Ray Trace analysis."""

    def __init__(
        self,
        *,
        hx: float = 0,
        hy: float = 0,
        px: float = 0,
        py: float = 0,
        wavelength: int = 1,
        field: int = 0,
        raytrace_type: constants.Analysis.Settings.Aberrations.RayTraceType | str = "DirectionCosines",
        global_coordinates: bool = False,
    ):
        """Create a new Single Ray Trace analysis.

        See Also
        --------
        SingleRayTraceSettings : Settings for the Single Ray Trace analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> SingleRayTraceResult:
        """Run the Single Ray Trace analysis."""
        self.analysis.Settings.Hx = self.settings.hx
        self.analysis.Settings.Hy = self.settings.hy
        self.analysis.Settings.Px = self.settings.px
        self.analysis.Settings.Py = self.settings.py
        self.analysis.set_wavelength(self.settings.wavelength)
        self.analysis.set_field(self.settings.field)
        self.analysis.Settings.Type = constants.process_constant(
            constants.Analysis.Settings.Aberrations.RayTraceType, self.settings.raytrace_type
        )
        self.analysis.Settings.UseGlobal = self.settings.global_coordinates

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Get results
        return self.parse_output(
            "single_ray_trace", transformer=SingleRayTraceTransformer, result_type=SingleRayTraceResult
        )



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("NSCSagRemoveOptions", "NSCSagShowAs")

class NSCSagRemoveOptions:
    None_: NSCSagRemoveOptions = None
    BaseROC: NSCSagRemoveOptions = None
    BestFitSphere: NSCSagRemoveOptions = None
    AverageSag: NSCSagRemoveOptions = None
    MinimumSag: NSCSagRemoveOptions = None

class NSCSagShowAs:
    Surface: NSCSagShowAs = None
    Contour: NSCSagShowAs = None
    GreyScale: NSCSagShowAs = None
    InverseGreyScale: NSCSagShowAs = None
    FalseColor: NSCSagShowAs = None
    InverseFalseColor: NSCSagShowAs = None
    X_CrossSection: NSCSagShowAs = None
    Y_CrossSection: NSCSagShowAs = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("FileSource", "KnownFolder", "Operation")

class FileSource:
    Project: FileSource = None
    Root: FileSource = None
    Unknown: FileSource = None

class KnownFolder:
    Project: KnownFolder = None
    ABgData: KnownFolder = None
    BlackBoxes: KnownFolder = None
    Coatings: KnownFolder = None
    Configs: KnownFolder = None
    DllBulkScatter: KnownFolder = None
    DllDiffractive: KnownFolder = None
    DllGradientIndex: KnownFolder = None
    DllObjects: KnownFolder = None
    DllPhysicalOptics: KnownFolder = None
    DllSources: KnownFolder = None
    DllSurfaceScatter: KnownFolder = None
    DllSurfaces: KnownFolder = None
    Extend: KnownFolder = None
    GlassCat: KnownFolder = None
    ImaFiles: KnownFolder = None
    Macros: KnownFolder = None
    MeritFunction: KnownFolder = None
    Miscellaneous: KnownFolder = None
    ObjectsApertures: KnownFolder = None
    ObjectsCadFiles: KnownFolder = None
    ObjectsCreoParametricFiles: KnownFolder = None
    ObjectsGridFiles: KnownFolder = None
    ObjectsInventorFiles: KnownFolder = None
    ObjectsPartDesignerObjects: KnownFolder = None
    ObjectsPhosphorsAndFluorescenceFiles: KnownFolder = None
    ObjectsPolygonObjects: KnownFolder = None
    ObjectsStopFiles: KnownFolder = None
    ObjectsSolidWorksFiles: KnownFolder = None
    ObjectsSourcesEulumdat: KnownFolder = None
    ObjectsSourcesIesna: KnownFolder = None
    ObjectsSourcesRadiantSourceModelFiles: KnownFolder = None
    ObjectsSourcesSourceFiles: KnownFolder = None
    ObjectsSourcesSpectrumFiles: KnownFolder = None
    ObjectsTabulatedObjects: KnownFolder = None
    PopBeamFiles: KnownFolder = None
    Profiles: KnownFolder = None
    ScatterData: KnownFolder = None
    Tolerance: KnownFolder = None
    Udo: KnownFolder = None
    ZosApiExtensions: KnownFolder = None
    ZosApiOperands: KnownFolder = None
    ZosApiUserAnalysis: KnownFolder = None
    DllBirefringence: KnownFolder = None
    Unknown: KnownFolder = None

class Operation:
    None_: Operation = None
    GetFilesInUse: Operation = None
    GetAdditionalFiles: Operation = None



"""Base functionality for analyses.

This module provides classes and functions to handle the creation, configuration, and execution of analyses, as well as
the serialization and deserialization of analysis results.

Classes
-------
AnalysisMessage
    Represents a message from a OpticStudio analysis.
AnalysisMetadata
    Contains metadata from a OpticStudio analysis.
AnalysisResult
    Represents the result of a OpticStudio analysis.
OnComplete
    Enum defining actions to perform after running an analysis.
Analysis
    Provides access to the settings and results of a OpticStudio analysis.
AnalysisWrapper
    Abstract base class for analysis wrappers.

Functions
---------
new_analysis(oss, analysis_type, settings_first=True)
    Creates a new analysis in OpticStudio.
"""

from __future__ import annotations

import dataclasses
import os
import weakref
from abc import ABC, abstractmethod
from dataclasses import dataclass, is_dataclass
from datetime import datetime  # noqa: TCH003 Pydantic needs datetime to be present at runtime
from enum import Enum
from importlib import import_module
from pathlib import Path
from tempfile import mkstemp
from typing import TYPE_CHECKING, Generic, Literal, TypedDict, TypeVar, cast, get_args

import numpy as np
import pandas as pd
import pydantic
from pydantic import (
    BaseModel,
    ConfigDict,
    RootModel,
    TypeAdapter,
    field_serializer,
    model_serializer,
    model_validator,
)

from zospy.analyses.parsers import load_grammar, parse
from zospy.analyses.parsers.types import ValidatedDataFrame
from zospy.api import _ZOSAPI, constants
from zospy.utils import zputils
from zospy.utils.clrutils import system_datetime_to_datetime

if TYPE_CHECKING:
    import sys

    from lark import Transformer
    from pydantic_core.core_schema import SerializerFunctionWrapHandler

    from zospy.zpcore import OpticStudioSystem

    if sys.version_info <= (3, 11):
        from typing_extensions import NotRequired
    else:
        from typing import NotRequired


__all__ = (
    "Analysis",
    "AnalysisResult",
    "BaseAnalysisWrapper",
    "OnComplete",
    "new_analysis",
)


@dataclass(frozen=True)
class AnalysisMessage:
    """Zemax OpticStudio analysis message."""

    ErrorCode: str
    Message: str


@dataclass(frozen=True)
class AnalysisMetadata:
    """Zemax OpticStudio analysis metadata."""

    DateTime: datetime
    FeatureDescription: str
    LensFile: str
    LensTitle: str


AnalysisData = TypeVar("AnalysisData")
AnalysisSettings = TypeVar("AnalysisSettings")


class _TypeInfo(TypedDict):
    data_type: Literal["dataframe", "ndarray", "zospy_class", "none"]
    name: NotRequired[str | None]
    module: NotRequired[str | None]


def _serialize_analysis_data_type(data: AnalysisData) -> _TypeInfo:
    if data is None:
        return {"data_type": "none"}

    if isinstance(data, pd.DataFrame):
        return {"data_type": "dataframe"}

    if isinstance(data, np.ndarray):
        return {"data_type": "ndarray"}

    if is_dataclass(data) or isinstance(data, BaseModel):
        return {"data_type": "zospy_class", "name": type(data).__name__, "module": type(data).__module__}

    raise ValueError(f"Cannot serialize data type: {type(data)}")


def _deserialize_zospy_class(data: dict, typeinfo: _TypeInfo) -> AnalysisData:
    if typeinfo["module"].startswith("zospy.analyses"):
        try:
            m = import_module(typeinfo["module"])
            t = getattr(m, typeinfo["name"])

            return TypeAdapter(t).validate_python(data)
        except (ModuleNotFoundError, AttributeError):
            return data

    return data


def _deserialize_analysis_data(data: dict | list, typeinfo: _TypeInfo) -> AnalysisData:
    if typeinfo["data_type"] == "none":
        return None

    if typeinfo["data_type"] == "dataframe":
        return pd.DataFrame.from_dict(data, orient="tight")

    if typeinfo["data_type"] == "ndarray":
        return np.array(data)

    if typeinfo["data_type"] == "zospy_class":
        return _deserialize_zospy_class(data, typeinfo)

    raise ValueError(f"Cannot deserialize data type: {typeinfo['data_type']}")


@pydantic.dataclasses.dataclass(frozen=True, config=ConfigDict(ser_json_inf_nan="constants"))
class AnalysisResult(Generic[AnalysisData, AnalysisSettings]):
    """Zemax OpticStudio analysis result.

    Attributes
    ----------
    data : AnalysisData
        The data of the analysis. Can be a `pandas.DataFrame`, `numpy.ndarray`, or an analysis-specific dataclass.
    settings : AnalysisSettings
        The settings of the analysis.
    metadata : AnalysisMetadata
        Metadata of the analysis. Contains the date and time of the analysis, the feature description, the lens file,
        and the lens title.
    header : list[str] | None
        The header data of the analysis. Only available for some analyses.
    messages : list[AnalysisMessage] | None
        Error messages from the analysis.
    """

    data: AnalysisData
    settings: AnalysisSettings
    metadata: AnalysisMetadata
    header: list[str] | None = None
    messages: list[AnalysisMessage] | None = None

    def to_json(self):
        """Convert the result to a JSON string."""
        return RootModel(self).model_dump_json(indent=4)

    @classmethod
    def from_json(cls, data: str):
        """Create a result from a JSON string."""
        return TypeAdapter(cls).validate_json(data)

    @field_serializer("data", mode="wrap", when_used="json")
    def _serialize_data(
        self,
        value: AnalysisData,
        nxt: SerializerFunctionWrapHandler,
        info,  # noqa: ARG002
    ):
        if isinstance(value, pd.DataFrame):
            return TypeAdapter(ValidatedDataFrame, config=ConfigDict(ser_json_inf_nan="constants")).dump_python(
                value, mode="json"
            )

        if isinstance(value, np.ndarray):
            return value.tolist()

        return nxt(value)

    @model_serializer(mode="wrap", when_used="json")
    def _serialize_types(self, nxt: SerializerFunctionWrapHandler):
        data = nxt(self)
        data["__analysis_data__"] = _serialize_analysis_data_type(self.data)
        data["__analysis_settings__"] = {
            "data_type": "zospy_class",
            "name": type(self.settings).__name__,
            "module": type(self.settings).__module__,
        }

        return data

    @model_validator(mode="wrap")
    @classmethod
    def _deserialize_types(cls, data: any, handler):
        if isinstance(data, dict):
            if "__analysis_data__" in data:
                data["data"] = _deserialize_analysis_data(data["data"], data.pop("__analysis_data__"))
            if "__analysis_settings__" in data:
                data["settings"] = _deserialize_zospy_class(data["settings"], data.pop("__analysis_settings__"))

        return handler(data)


class OnComplete(str, Enum):
    """Action to perform after running an OpticStudio analysis.

    This `Enum` can be passed to analysis methods in `zospy.analyses`.

    Attributes
    ----------
    Close : str
        Close the analysis.
    Release : str
        Keep the analysis open, but not active.
    Sustain : str
        Keep the analysis open and active.

    Examples
    --------
    Run a System Data analysis and keep the analysis open:

    >>> from zospy.analyses import OnComplete
    >>> from zospy.analyses.reports import system_data
    >>> result = system_data(oss, OnComplete.Sustain)
    """

    Close = "Close"
    """Close the analysis."""

    Release = "Release"
    """Keep the analysis open, but not active."""

    Sustain = "Sustain"
    """Keep the analysis open and active."""


class _ValidatedSetter:
    """Wrapper class that only allows to set existing properties."""

    __slots__ = ("_obj",)

    def __init__(self, obj):
        self._obj = obj

    def __getattr__(self, name):
        return getattr(self._obj, name)

    def __setattr__(self, name, value):
        if name in self.__slots__:
            super().__setattr__(name, value)
        elif hasattr(self._obj, name):
            setattr(self._obj, name, value)
        else:
            raise AttributeError(f"'{type(self._obj).__name__}' object has no attribute '{name}'")


_ValidatedSetterType = TypeVar("_ValidatedSetterType")


def _validated_setter(obj: _ValidatedSetterType) -> _ValidatedSetterType:
    """Wrap an object to only allow setting existing properties.

    Helper function that retains the original object's type information.

    Parameters
    ----------
    obj : Any
        The object to wrap.
    """
    return _ValidatedSetter(obj)


class Analysis:
    """OpticStudio analysis.

    This class wraps ZOSAPI analysis objects to provide direct access to its settings and results.
    All properties and methods of the ZOSAPI analysis object are available through this class.
    """

    def __init__(self, analysis: _ZOSAPI.Analysis.IA_):
        """Zemax OpticStudio Analysis with full access to its Settings and Results attributes.

        Parameters
        ----------
        analysis : ZOSAPI.Analysis.IA_
            analysis object
        """
        self._analysis = _validated_setter(analysis)

    @property
    def Settings(self) -> _ZOSAPI.Analysis.Settings.IAS_:  # noqa: N802
        """Analysis-specific settings."""
        return _validated_setter(self._analysis.GetSettings())

    @property
    def Results(self) -> _ZOSAPI.Analysis.Data.IAR_:  # noqa: N802
        """Analysis results."""
        return self._analysis.GetResults()

    def get_field(self) -> int | Literal["All"]:
        """Get the field value from the analysis settings.

        Returns
        -------
        int | Literal["All"]
            Either the field number, or 'All' if field was set to 'All'.
        """
        field_number = self.Settings.Field.GetFieldNumber()

        return "All" if field_number == 0 else field_number

    def set_field(self, value: int | Literal["All"]):
        """Set the field value in the analysis settings.

        Parameters
        ----------
        value : int | Literal["All"]
            The value to which the field should be set. Either int or str. Accepts only 'All' as string.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            When 'value' is not integer or string. When it is a string, it also raises an error when the string
            does not equal 'All'.
            When 'value' is not accepted by the analysis as field specification.

        Warnings
        --------
        The ZOS-API handles invalid integer values without warning about this:

        - If selecting all fields is not possible, a value of `0` will instead select the first field;
        - If the value is greater than the number of fields, the last field will be selected.
        """
        if value == "All":
            message = self.Settings.Field.UseAllFields()
        elif isinstance(value, int):
            message = self.Settings.Field.SetFieldNumber(value)
        else:
            raise ValueError(f'Field value should be "All" or an integer, got {value}')

        if message is not None:
            raise ValueError(f"Could not set field value to {value}: {message.Text}")

    @property
    def header_data(self) -> list[str]:
        """Obtain the header data from an OpticStudio analysis.

        Returns
        -------
        list
            The header data.
        """
        return list(self.Results.HeaderData.Lines)

    @property
    def messages(self) -> list[AnalysisMessage]:
        """Obtain the messages from the analysis.

        Returns
        -------
        list[AnalysisMessage]
            A list of messages. Each item is a NamedTuple with names (ErrorCode, Message).
        """
        result = []

        for i in range(self.Results.NumberOfMessages):
            message = self.Results.GetMessageAt(i)
            error_code = str(message.ErrorCode)

            result.append(AnalysisMessage(error_code, message.Text))

        return result

    @property
    def metadata(self) -> AnalysisMetadata:
        """Obtain the metadata from the analysis.

        Returns
        -------
        AnalysisMetadata
            A named tuple containing the MetaData including 'DateTime', 'FeatureDescription', 'LensFile'
            and 'LensTitle'.
        """
        return AnalysisMetadata(
            system_datetime_to_datetime(self.Results.MetaData.Date),
            self.Results.MetaData.FeatureDescription,
            self.Results.MetaData.LensFile,
            self.Results.MetaData.LensTitle,
        )

    def get_wavelength(self) -> int | Literal["All"]:
        """Get the wavelength value from the analysis settings.

        Returns
        -------
        int | Literal["All"]
            Either the wavelength number, or 'All' if wavelength was set to 'All'.
        """
        wavelength = self.Settings.Wavelength.GetWavelengthNumber()

        return "All" if wavelength == 0 else wavelength

    def set_wavelength(self, value: int | Literal["All"]):
        """Set the wavelength value in the analysis settings.

        Parameters
        ----------
        value : int | Literal["All"]
            The value to which the wavelength should be set. Either int or str. Accepts only 'All' as string.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            When 'value' is not integer or string. When it is a string, it also raises an error when the string does not
            equal 'All'.
            When 'value' is not accepted by the analysis as wavelength specification.

        Warnings
        --------
        The ZOS-API handles invalid integer values without warning about this:

        - If selecting all wavelengths is not possible, a value of `0` will instead select the first wavelength;
        - If the value is greater than the number of wavelengths, the last wavelength will be selected.
        """
        if value == "All":
            message = self.Settings.Wavelength.UseAllWavelengths()
        elif isinstance(value, int):
            message = self.Settings.Wavelength.SetWavelengthNumber(value)
        else:
            raise ValueError('Wavelength value should be "All" or an integer')

        if message is not None:
            raise ValueError(f"Could not set wavelength value to {value}: {message.Text}")

    def set_surface(self, value: int | Literal["Image", "Objective"]):
        """Set the surface value in the analysis settings.

        Parameters
        ----------
        value : int | Literal["Image", "Objective"]
            The value to which the surface should be set. Either int or str. Accepts only 'Image' or 'Objective' as
            string.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            When 'value' is not integer or string. When it is a string, it also raises an error when the string does not
            equal 'Image' or 'Objective'.
            When 'value' is not accepted by the analysis as surface specification.

        Warnings
        --------
        The ZOS-API handles invalid integer values without warning about this:

        - If the supplied value is `0`, OpticStudio will base it selection based on the analysis. In some, it will
          select surface 0. In others, it will select a special surface (e.g. 'Image'). If neither is available, it
          selects the first available surface;
        - If the value is greater than the number of surfaces, the last surface will be selected.
        """
        if value == "Image":
            message = self.Settings.Surface.UseImageSurface()
        elif value == "Objective":
            message = self.Settings.Surface.UseObjectiveSurface()
        elif isinstance(value, int):
            message = self.Settings.Surface.SetSurfaceNumber(value)
        else:
            raise ValueError(f'Surface value should be "Image", "Objective" or an integer, got {value}')

        if message is not None:
            raise ValueError(f"Could not set surface value to {value}: {message.Text}")

    def get_text_output(self, txtoutfile: str, encoding: str):
        """Get the text output of the analysis.

        Parameters
        ----------
        txtoutfile : str
            Path to the text output file. The file will be created if it does not exist.
        encoding : str
            The encoding of the text file. The encoding used by OpticStudio can be obtained with
            `zospy.ZOS.get_txtfile_encoding`.
        """
        self.Results.GetTextFile(txtoutfile)

        with open(txtoutfile, encoding=encoding) as f:
            return f.read()

    def __getattr__(self, item):
        """Get an attribute from the analysis object.

        If the attribute is not found, it is retrieved from the ZOSAPI analysis object.
        """
        return getattr(self._analysis, item)

    def __dir__(self):
        """List the attributes of the Analysis wrapper and the ZOSAPI analysis object."""
        return sorted(set(super().__dir__()).union(dir(self._analysis)))


def new_analysis(
    oss: OpticStudioSystem,
    analysis_type: constants.Analysis.AnalysisIDM,
    *,
    settings_first: bool = True,
) -> Analysis:
    """Create a new analysis in OpticStudio.

    Parameters
    ----------
    oss : OpticStudioSystem
        The Zemax OpticStudio system
    analysis_type : zospy.constants.Analysis.AnalysisIDM
        Analysis type from `ZOSAPI.Analysis.AnalysisIDM`
    settings_first : bool
        Do not run the analysis immediately, which allows to adjust settings. Defaults to `True`.

    Returns
    -------
    analysis : Analysis

    Examples
    --------
    >>> import zospy as zp
    >>> new_analysis(oss, zp.constants.Analysis.AnalysisIDM.ZernikeStandardCoefficients)
    """
    analysis = (
        oss.Analyses.New_Analysis_SettingsFirst(analysis_type)
        if settings_first
        else oss.Analyses.New_Analysis(analysis_type)
    )
    return Analysis(analysis)


class BaseAnalysisWrapper(ABC, Generic[AnalysisData, AnalysisSettings]):
    """Base class for analysis wrappers.

    This class provides a common interface for all analysis wrappers. It defines the methods and properties that all
    analysis wrappers should implement.

    Attributes
    ----------
    TYPE : str
        The type of the analysis. This should match the `AnalysisIDM` constant in OpticStudio.
    _needs_config_file : bool
        Flag to indicate if the analysis requires a configuration file.
    _needs_text_output_file : bool
        Flag to indicate if the analysis requires a text output file.
    """

    TYPE: str = None
    MODE: Literal["Sequential", "Nonsequential"] | None = None

    # Flags to indicate if the analysis needs a configuration file or text output file
    _needs_config_file: bool = False
    _needs_text_output_file: bool = False

    def __init__(self, *, settings_kws: dict[str, any] | None = None):
        """Create a new analysis wrapper.

        Settings can be changed by passing the settings as keyword arguments. Use the `with_settings` method to specify
        the settings using a settings object.

        Parameters
        ----------
        settings_kws : dict[str, any]
            Arguments to set the settings of the analysis.

        Raises
        ------
        ValueError
            If `settings` is not a dataclass.
        """
        self._settings = self._default_settings()
        self.update_settings(settings_kws=settings_kws)

        self._config_file = None
        self._text_output_file = None

        self._oss = None
        self._analysis = None
        self._remove_config_file = False
        self._remove_text_output_file = False

    def __init_subclass__(
        cls,
        *,
        analysis_type: str | None = None,
        mode: Literal["Sequential", "Nonsequential"] | None = None,
        needs_config_file: bool = False,
        needs_text_output_file: bool = False,
        **kwargs,
    ):
        """Determine the settings type and class-level configuration of the analysis."""
        cls.TYPE = analysis_type
        cls.MODE = mode
        cls._needs_config_file = needs_config_file
        cls._needs_text_output_file = needs_text_output_file

        if not hasattr(cls, "_settings_type"):
            if hasattr(cls, "__orig_bases__"):
                base = cls.__orig_bases__[0]
                cls._settings_type: type[AnalysisSettings] = get_args(base)[1]
            else:
                cls._settings_type = type(None)  # TODO: change to NoneType when dropping support for Python 3.9

        super().__init_subclass__(**kwargs)

    def update_settings(
        self,
        *,
        settings: AnalysisSettings | None = None,
        settings_kws: dict[str, any] | None = None,
    ) -> None:
        """Update the settings of the analysis using a settings object or keyword arguments.

        Settings can be specified as an object and as keyword arguments. If both are specified, the keyword arguments
        take precedence. If no settings are specified, the default settings are used. Furthermore, instead of using
        a reference to the settings object, a new settings object is created with the specified parameters. This is done
        to avoid modifying the original settings object.

        Parameters
        ----------
        settings : AnalysisSettings
            Analysis settings object.
        settings_kws
            Dictionary with the settings parameters.

        Raises
        ------
        ValueError
            If `settings` is not a dataclass.
        """
        # Use the existing settings if no settings are specified
        settings = settings or self.settings

        if settings is None:
            # Analysis does not have settings
            return

        if not is_dataclass(settings):
            raise TypeError("settings should be a dataclass.")

        # Create a new settings object with the specified parameters. If no parameters are specified, this creates a
        # copy of the settings object. This is done to avoid modifying the original settings object.
        self._settings = dataclasses.replace(settings, **(settings_kws or {}))

    @classmethod
    def _default_settings(cls) -> AnalysisSettings:
        """Get the default settings of the analysis.

        Returns
        -------
        AnalysisSettings
            The default settings.
        """
        return cls._settings_type() if cls._settings_type is not None else None

    @classmethod
    def with_settings(cls, settings: AnalysisSettings):
        """Create a new analysis with the specified settings.

        Parameters
        ----------
        settings : AnalysisSettings
            Settings of the analysis.

        Returns
        -------
        BaseAnalysisWrapper
            The analysis wrapper.
        """
        instance = cls()
        instance.update_settings(settings=settings)

        return instance

    @property
    def settings(self) -> AnalysisSettings:
        """Settings of the analysis."""
        return self._settings

    @property
    def config_file(self) -> Path | None:
        """Path to the temporary configuration file.

        This file is used to store the settings of the analysis and can only be set from the `run` method.

        Returns
        -------
        Path | None
            Path to the temporary configuration file.

        Raises
        ------
        ValueError
            If the analysis does not require a configuration file.
        """
        if not self._needs_config_file:
            raise ValueError("This analysis does not require a configuration file.")

        return self._config_file

    @property
    def text_output_file(self) -> Path | None:
        """Path to the temporary text output file.

        This file is used to store the text output of the analysis and can only be set from the `run` method.

        Returns
        -------
        Path | None
            Path to the temporary text output file.

        Raises
        ------
        ValueError
            If the analysis does not require a text output file.
        """
        if not self._needs_text_output_file:
            raise ValueError("This analysis does not require a text output file.")

        return self._text_output_file

    @property
    def oss(self) -> OpticStudioSystem:
        """The OpticStudio system. This property is set when the analysis is run."""
        if self._oss is None:
            raise ValueError("OpticStudioSystem has not been set.")

        return self._oss

    @property
    def analysis(self) -> Analysis | None:
        """The OpticStudio analysis object. This property is set when the analysis is run."""
        return self._analysis

    def get_text_output(self) -> str:
        """Get the text output of the analysis."""
        self.analysis.Results.GetTextFile(str(self.text_output_file))

        with open(self._text_output_file, encoding=self.oss.ZOS.get_txtfile_encoding()) as f:
            return f.read()

    def _create_analysis(self, *, settings_first=True):
        if self.analysis is not None and str(self.analysis.AnalysisType) == self.TYPE:
            return

        analysis_type = constants.process_constant(constants.Analysis.AnalysisIDM, self.TYPE)
        self._analysis = new_analysis(self.oss, analysis_type, settings_first=settings_first)

    @abstractmethod
    def run_analysis(self, *args, **kwargs) -> AnalysisData:
        """Run the analysis and return the results."""

    def _check_mode(self):
        if self.MODE is None:
            return

        if self.oss.Mode != self.MODE:
            raise ValueError(f"The analysis requires {self.MODE} mode, got {self.oss.Mode}.")

    @staticmethod
    def _create_tempfile(path: Path | None, suffix: str) -> (Path, bool):
        if path is None:
            fd, path = mkstemp(suffix=suffix, prefix="zospy_")
            os.close(fd)
            clean_file = True
        else:
            if not path.suffix == suffix:
                raise ValueError(f"File path should end with '{suffix}'.")

            clean_file = False

        return Path(path), clean_file

    def _complete(self, oncomplete: OnComplete = OnComplete.Close) -> None:
        """Completes the analysis by either closing, releasing or sustaining it."""
        if oncomplete == OnComplete.Close:
            self.analysis.Close()
            self._analysis = None
        elif oncomplete == OnComplete.Release:
            self.analysis.Release()
            self._analysis = None
        elif oncomplete == OnComplete.Sustain:
            return
        else:
            raise ValueError(f"oncomplete should be a member of zospy.analyses.base.OnComplete, got {oncomplete}")

    def run(
        self,
        oss: OpticStudioSystem,
        config_file: str | Path | None = None,
        text_output_file: str | Path | None = None,
        oncomplete: OnComplete | Literal["Close", "Release", "Sustain"] = "Close",
    ) -> AnalysisResult[AnalysisData, AnalysisSettings]:
        """Run the analysis and return the results.

        This method opens the analysis in OpticStudio and creates temporary files if needed. After running the analysis,
        the temporary files are removed and the analysis is closed, released, or sustained based on `oncomplete`.

        The `config_file` is ignored if self._needs_config_file is `False`.
        The `text_output_file` is ignored if self._needs_text_output_file is `False`.

        Parameters
        ----------
        oss : OpticStudioSystem
            The OpticStudio system.
        config_file : str | Path | None
            Path to the configuration file. If `None`, a temporary file will be created.
        text_output_file : str | Path | None
            Path to the text output file. If `None`, a temporary file will be created.
        oncomplete : OnComplete | Literal["Close", "Release", "Sustain"]
            Action to perform after running the analysis. If "Close", the analysis will be closed. If "Release", the
            analysis will be kept open but not active. If "Sustain", the analysis will be kept open and active.

        Returns
        -------
        AnalysisResult
            The analysis results.
        """
        self._oss = weakref.proxy(oss)
        self._check_mode()
        self._create_analysis()

        if self._needs_config_file:
            self._config_file, self._remove_config_file = self._create_tempfile(config_file, ".CFG")

        if self._needs_text_output_file:
            self._text_output_file, self._remove_text_output_file = self._create_tempfile(text_output_file, ".txt")

        data = self.run_analysis()

        result = AnalysisResult(
            data,
            settings=self.settings,
            metadata=self.analysis.metadata,
            header=self.analysis.header_data,
            messages=self.analysis.messages,
        )

        if self._remove_config_file:
            os.remove(self._config_file)

        if self._remove_text_output_file:
            os.remove(self._text_output_file)

        self._complete(OnComplete(oncomplete))

        return result

    def parse_output(
        self,
        grammar: str,
        transformer: type[Transformer],
        result_type: type[AnalysisData],
    ) -> AnalysisData:
        """Parse the text output of the analysis."""
        parser = load_grammar(grammar)
        parse_result = parse(self.get_text_output(), parser, transformer)

        return cast(result_type, result_type(**parse_result))

    @staticmethod
    def _process_data_series_or_grid(data: list[pd.DataFrame]) -> pd.DataFrame | None:
        if len(data) == 0:
            return None

        if len(data) == 1:
            return data[0]

        return pd.concat(data, axis=1)

    def get_data_series(self) -> pd.DataFrame | None:
        """Get the data series from the analysis result.

        Returns
        -------
        pd.Series | None
            The data series from the analysis result, or None if there are no data series.
        """
        data = [
            zputils.unpack_dataseries(self.analysis.Results.DataSeries[i])
            for i in range(self.analysis.Results.NumberOfDataSeries)
        ]

        return self._process_data_series_or_grid(data)

    def get_data_grid(self, cell_origin: Literal["bottom_left", "center"] = "bottom_left") -> pd.DataFrame | None:
        """Get the data grids from the analysis result.

        Parameters
        ----------
        cell_origin : Literal["bottom_left", "center"]
            Defines how minx and miny are handled to determine coordinates. Either 'bottom_left' indicating that they
            are defining the bottom left of the grd cell, or 'center', indicating that they provide the center of the
            grid cell. Defaults to 'bottom_left'.

        Returns
        -------
        pd.DataFrame | None
            The data grids from the analysis result, or None if there are no data grids.
        """
        data = [
            zputils.unpack_datagrid(self.analysis.Results.DataGrids[i], cell_origin=cell_origin)
            for i in range(self.analysis.Results.NumberOfDataGrids)
        ]

        return self._process_data_series_or_grid(data)

    def __call__(self, oss: OpticStudioSystem, *args, **kwargs):
        """Run the analysis and return the results."""
        return self.run(oss, *args, **kwargs)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "FieldColumn",
    "FieldNormalizationType",
    "FieldPattern",
    "FieldType",
    "FNumberComputationType",
    "HuygensIntegralSettings",
    "ParaxialRaysSetting",
    "PolarizationMethod",
    "QuadratureSteps",
    "RayAimingMethod",
    "RayAimingType",
    "ReferenceOPDSetting",
    "WavelengthPreset",
    "ZemaxAfocalModeUnits",
    "ZemaxAnalysisUnits",
    "ZemaxApertureType",
    "ZemaxApodizationType",
    "ZemaxMTFUnits",
    "ZemaxSourceUnits",
    "ZemaxSystemUnits",
    "ZemaxUnitPrefix",
)

class FieldColumn:
    Comment: FieldColumn = None
    X: FieldColumn = None
    Y: FieldColumn = None
    Weight: FieldColumn = None
    VDX: FieldColumn = None
    VDY: FieldColumn = None
    VCX: FieldColumn = None
    VCY: FieldColumn = None
    TAN: FieldColumn = None
    VAN: FieldColumn = None

class FieldNormalizationType:
    Radial: FieldNormalizationType = None
    Rectangular: FieldNormalizationType = None

class FieldPattern:
    UniformY: FieldPattern = None
    EqualAreaY: FieldPattern = None
    UniformX: FieldPattern = None
    EqualAreaX: FieldPattern = None
    Grid: FieldPattern = None
    UniformRadial: FieldPattern = None
    EqualAreaRadial: FieldPattern = None

class FieldType:
    Angle: FieldType = None
    ObjectHeight: FieldType = None
    ParaxialImageHeight: FieldType = None
    RealImageHeight: FieldType = None
    TheodoliteAngle: FieldType = None

class FNumberComputationType:
    TracingRays: FNumberComputationType = None
    PupilSizePosition: FNumberComputationType = None

class HuygensIntegralSettings:
    Auto: HuygensIntegralSettings = None
    Planar: HuygensIntegralSettings = None
    Spherical: HuygensIntegralSettings = None

class ParaxialRaysSetting:
    ConsiderCoordinateBreaks: ParaxialRaysSetting = None
    IgnoreCoordinateBreaks: ParaxialRaysSetting = None

class PolarizationMethod:
    XAxisMethod: PolarizationMethod = None
    YAxisMethod: PolarizationMethod = None
    ZAxisMethod: PolarizationMethod = None

class QuadratureSteps:
    S2: QuadratureSteps = None
    S4: QuadratureSteps = None
    S6: QuadratureSteps = None
    S8: QuadratureSteps = None
    S10: QuadratureSteps = None
    S12: QuadratureSteps = None

class RayAimingMethod:
    Off: RayAimingMethod = None
    Paraxial: RayAimingMethod = None
    Real: RayAimingMethod = None

class RayAimingType:
    Heuristic: RayAimingType = None
    Optimize: RayAimingType = None

class ReferenceOPDSetting:
    Absolute: ReferenceOPDSetting = None
    Infinity: ReferenceOPDSetting = None
    ExitPupil: ReferenceOPDSetting = None
    Absolute2: ReferenceOPDSetting = None

class WavelengthPreset:
    FdC_Visible: WavelengthPreset = None
    Photopic_Bright: WavelengthPreset = None
    Scotopic_Dark: WavelengthPreset = None
    HeNe_0p6328: WavelengthPreset = None
    HeNe_0p5438: WavelengthPreset = None
    Argon_0p4880: WavelengthPreset = None
    Argon_0p5145: WavelengthPreset = None
    NDYAG_1p0641: WavelengthPreset = None
    NDGlass_1p054: WavelengthPreset = None
    CO2_10p60: WavelengthPreset = None
    CrLiSAF_0p840: WavelengthPreset = None
    TiAl203_0p760: WavelengthPreset = None
    Ruby_0p6943: WavelengthPreset = None
    HeCadmium_0p4416: WavelengthPreset = None
    HeCadmium_0p3536: WavelengthPreset = None
    HeCadmium_0p3250: WavelengthPreset = None
    t_1p014: WavelengthPreset = None
    r_0p707: WavelengthPreset = None
    C_0p656: WavelengthPreset = None
    d_0p587: WavelengthPreset = None
    F_0p486: WavelengthPreset = None
    g_0p436: WavelengthPreset = None
    i_0p365: WavelengthPreset = None
    Fp_0p365: WavelengthPreset = None
    e_0p54607: WavelengthPreset = None
    Cp_0p6438469: WavelengthPreset = None
    FpeCp_Visible: WavelengthPreset = None
    THz_193p10: WavelengthPreset = None

class ZemaxAfocalModeUnits:
    Microradians: ZemaxAfocalModeUnits = None
    Milliradians: ZemaxAfocalModeUnits = None
    Radians: ZemaxAfocalModeUnits = None
    ArcSeconds: ZemaxAfocalModeUnits = None
    ArcMinutes: ZemaxAfocalModeUnits = None
    Degrees: ZemaxAfocalModeUnits = None

class ZemaxAnalysisUnits:
    WattsPerMMSq: ZemaxAnalysisUnits = None
    WattsPerCMSq: ZemaxAnalysisUnits = None
    WattsPerinSq: ZemaxAnalysisUnits = None
    WattsPerMSq: ZemaxAnalysisUnits = None
    WattsPerftSq: ZemaxAnalysisUnits = None

class ZemaxApertureType:
    EntrancePupilDiameter: ZemaxApertureType = None
    ImageSpaceFNum: ZemaxApertureType = None
    ObjectSpaceNA: ZemaxApertureType = None
    FloatByStopSize: ZemaxApertureType = None
    ParaxialWorkingFNum: ZemaxApertureType = None
    ObjectConeAngle: ZemaxApertureType = None

class ZemaxApodizationType:
    Uniform: ZemaxApodizationType = None
    Gaussian: ZemaxApodizationType = None
    CosineCubed: ZemaxApodizationType = None

class ZemaxMTFUnits:
    CyclesPerMillimeter: ZemaxMTFUnits = None
    CyclesPerMilliradian: ZemaxMTFUnits = None

class ZemaxSourceUnits:
    Watts: ZemaxSourceUnits = None
    Lumens: ZemaxSourceUnits = None
    Joules: ZemaxSourceUnits = None

class ZemaxSystemUnits:
    Millimeters: ZemaxSystemUnits = None
    Centimeters: ZemaxSystemUnits = None
    Inches: ZemaxSystemUnits = None
    Meters: ZemaxSystemUnits = None

class ZemaxUnitPrefix:
    Femto: ZemaxUnitPrefix = None
    Pico: ZemaxUnitPrefix = None
    Nano: ZemaxUnitPrefix = None
    Micro: ZemaxUnitPrefix = None
    Milli: ZemaxUnitPrefix = None
    None_: ZemaxUnitPrefix = None
    Kilo: ZemaxUnitPrefix = None
    Mega: ZemaxUnitPrefix = None
    Giga: ZemaxUnitPrefix = None
    Tera: ZemaxUnitPrefix = None



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import importlib.metadata
from datetime import datetime
from pathlib import Path
from shutil import copytree

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "ZOSPy"
copyright = f"2023-{datetime.now().year}, Jan-Willem M. Beenakker, Luc van Vught, Corné Haasjes"  # noqa: A001, DTZ005
author = "Jan-Willem M. Beenakker, Luc van Vught, Corné Haasjes"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "nbsphinx",
    "numpydoc",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx_design",
]

myst_enable_extensions = ["colon_fence", "attrs_block", "attrs_inline", "substitution"]
myst_heading_anchors = 3
myst_substitutions = {
    "REFERENCE_VERSION": "2025 R1.01",
    "PYTHON_VERSIONS": "3.9 - 3.13",
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**/.conda", "**/.ipynb_checkpoints"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_theme_options = {
    "home_page_in_toc": True,
    "repository_url": "https://github.com/MREYE-LUMC/ZOSPy",
    "use_repository_button": True,
}

# -- Options for Sphinx autodoc and numpydoc ---------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration
# https://numpydoc.readthedocs.io/en/latest/install.html

# Display type hints in short form and only in the signature
autodoc_typehints = "signature"
autodoc_class_signature = "separated"
autodoc_typehints_format = "short"

# Mock imports that may be unavailable in the build environment (e.g. readthedocs)
autodoc_mock_imports = ["winreg", "clr", "System", "Python"]

numpydoc_class_members_toctree = False

# Do not document inherited class members for these types, e.g. because they extend built-in Python types
numpydoc_show_inherited_class_members = {
    "zospy.analyses.base.AttrDict": False,
    "zospy.analyses.base.AnalysisResult": False,
    "zospy.analyses.base.OnComplete": False,
}

# -- Options for nbsphinx (example notebooks) --------------------------------
# https://nbsphinx.readthedocs.io/
documentation_directory = Path(__file__).parent
example_directory = documentation_directory.parent / "examples"

# Copy examples to the documentation directory
for example in example_directory.iterdir():
    # Only include examples that are provided as notebooks
    if len(list(example.glob("*.ipynb"))) > 0:
        copytree(example, documentation_directory / "examples" / example.name, dirs_exist_ok=True)

# Add a banner to each example notebook included in the documentation
zp_version = importlib.metadata.version("zospy")
nbsphinx_prolog = rf"""
{{% set docname = env.doc2path(env.docname, base=None) %}}

.. raw:: html

    <div class="admonition note">
      This page was generated from a Jupyter notebook.
      <a href="https://github.com/MREYE-LUMC/ZOSPy/tree/v{zp_version}/examples/{{{{ env.docname.split('/')[-2] | e }}}}"
      class="reference external" download>Check the source code</a>
      or <a href="{{{{ env.docname.split('/') | last | e + '.ipynb' }}}}"
      class="reference download internal" download>download the notebook.</a>.
    </div>
"""

# -- Docstring preprocessing -------------------------------------------------
ANNOTATION_SUBSTITUTIONS = {"_ZOSAPI": "ZOSAPI"}


def apply_annotation_substitutions(app, obj: object, bound_method):  # noqa: ARG001
    """Substitute certain values in type annotations."""
    if not hasattr(obj, "__annotations__"):
        return

    for name, annotation in obj.__annotations__.items():
        if not isinstance(annotation, str):
            continue

        for old, new in ANNOTATION_SUBSTITUTIONS.items():
            if annotation.startswith(old):
                obj.__annotations__[name] = annotation.replace(old, new)


def setup(app):
    app.connect("autodoc-before-process-signature", apply_annotation_substitutions)



import pytest

from zospy.functions.nce import find_object_by_comment


@pytest.mark.parametrize(
    "comment,case_sensitive,expected_indices",
    [
        ("Abc", True, [1]),  # When using .ObjectNumber, the returned value matches the NCE numbering (starting at 1)
        ("Abc", False, [1, 2, 3]),
        ("ABc", True, [2, 3]),
        ("ABC", True, []),
        ("ABC", False, [1, 2, 3]),
        ("efg", True, []),
        ("efg", False, []),
    ],
)
def test_can_find_object_by_comment(nsc_simple_system, comment, case_sensitive, expected_indices):
    nsc_simple_system.NCE.GetObjectAt(1).Comment = "Abc"
    nsc_simple_system.NCE.GetObjectAt(2).Comment = "ABc"
    nsc_simple_system.NCE.GetObjectAt(3).Comment = "ABc"

    result = find_object_by_comment(nsc_simple_system.NCE, comment=comment, case_sensitive=case_sensitive)

    indices = [obj.ObjectNumber for obj in result]

    assert indices == expected_indices



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("SagittalAberrationComponent", "TangentialAberrationComponent")

class SagittalAberrationComponent:
    Aberration_X: SagittalAberrationComponent = None
    Aberration_Y: SagittalAberrationComponent = None

class TangentialAberrationComponent:
    Aberration_Y: TangentialAberrationComponent = None
    Aberration_X: TangentialAberrationComponent = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations
from . import LDE, MCE, MFE, NCE, TDE

__all__ = (
    "LDE",
    "MCE",
    "MFE",
    "NCE",
    "TDE",
    "CellDataType",
    "DirectionOfRayTravel",
    "EditorType",
    "ReflectTransmitCode",
    "SampleSides",
    "Samplings",
    "SolveStatus",
    "SolveType",
)

class CellDataType:
    Integer: CellDataType = None
    Double: CellDataType = None
    String: CellDataType = None

class DirectionOfRayTravel:
    inward: DirectionOfRayTravel = None
    outward: DirectionOfRayTravel = None

class EditorType:
    LDE: EditorType = None
    NCE: EditorType = None
    MFE: EditorType = None
    TDE: EditorType = None
    MCE: EditorType = None

class ReflectTransmitCode:
    Success: ReflectTransmitCode = None
    NoReflectDataInFile: ReflectTransmitCode = None
    NoTransmitDataInFile: ReflectTransmitCode = None

class SampleSides:
    Front: SampleSides = None
    Back: SampleSides = None

class Samplings:
    FiveDegrees: Samplings = None
    TwoDegrees: Samplings = None
    OneDegree: Samplings = None

class SolveStatus:
    Success: SolveStatus = None
    InvalidSolveType: SolveStatus = None
    InvalidRow: SolveStatus = None
    InvalidColumn: SolveStatus = None
    PostSurfaceStopOnly: SolveStatus = None
    InvalidMacro: SolveStatus = None
    Failed: SolveStatus = None

class SolveType:
    None_: SolveType = None
    Fixed: SolveType = None
    Variable: SolveType = None
    SurfacePickup: SolveType = None
    ZPLMacro: SolveType = None
    MarginalRayAngle: SolveType = None
    MarginalRayHeight: SolveType = None
    ChiefRayAngle: SolveType = None
    MarginalRayNormal: SolveType = None
    ChiefRayNormal: SolveType = None
    Aplanatic: SolveType = None
    ElementPower: SolveType = None
    CocentricSurface: SolveType = None
    ConcentricSurface: SolveType = None
    CocentricRadius: SolveType = None
    ConcentricRadius: SolveType = None
    FNumber: SolveType = None
    ChiefRayHeight: SolveType = None
    EdgeThickness: SolveType = None
    OpticalPathDifference: SolveType = None
    Position: SolveType = None
    Compensator: SolveType = None
    CenterOfCurvature: SolveType = None
    PupilPosition: SolveType = None
    MaterialSubstitute: SolveType = None
    MaterialOffset: SolveType = None
    MaterialModel: SolveType = None
    Automatic: SolveType = None
    Maximum: SolveType = None
    PickupChiefRay: SolveType = None
    ObjectPickup: SolveType = None
    ConfigPickup: SolveType = None
    ThermalPickup: SolveType = None
    MarginPercent: SolveType = None
    CA_fill: SolveType = None
    DIA_fill: SolveType = None
    DuplicateSag: SolveType = None
    InvertSag: SolveType = None
    FieldPickup: SolveType = None



"""OpticStudio analyses from the Extended Scene category."""

from zospy.analyses.extendedscene.geometric_image_analysis import (
    GeometricImageAnalysis,
    GeometricImageAnalysisSettings,
)

__all__ = ("GeometricImageAnalysis", "GeometricImageAnalysisSettings")



"""Geometric Image Analysis."""

from __future__ import annotations

from typing import Annotated, Literal, Union

import pandas as pd
from pandas import DataFrame
from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_settings
from zospy.analyses.parsers.types import WavelengthNumber, ZOSAPIConstant  # noqa: TCH001
from zospy.api import constants

__all__ = ("GeometricImageAnalysis", "GeometricImageAnalysisSettings")

from zospy.utils import zputils


@analysis_settings
class GeometricImageAnalysisSettings:
    """Settings for the Geometric Image Analysis.

    For an in depth explanation of the parameters, see the OpticStudio user manual.

    Attributes
    ----------
    wavelength : str | int
        The wavelength to use in the analysis. Either 'All' or an integer specifying the wavelength number. Defaults to
        'All'.
    field : int
        The field to use in the analysis. Should be an integer specifying the field number. Defaults to 1.
    surface : str | int
        The surface that is to be analyzed. Either 'Image', 'Object' or an integer. Defaults to 'Image'.
    field_size : float
        The width of the image in field coordinates. Defaults to 0.
    image_size : float
        When show_as is set to 'SpotDiagram', this defines the size of the scale bar superimposed on the image. Else, it
         defines the size of the detector used to capture the rays. Defaults to 50.
    file : str
        The image file used for the analysis. Must be an .IMA of .BIM file and reside in the OpticStudio images folder.
        Defaults to 'LETTERF.IMA'.
    rotation : float
        The rotation angle in degrees. Defaults to 0.
    rays_x_1000 : int
        The approximate amount of rays that are traced. See OpticStudio manual for further clarification. Defaults to
        10.
    show_as : str | constants.Analysis.GiaShowAsTypes
        Defines how the result is displayed in OpticStudio. Defaults to 'Surface'. Note that 'SpotDiagram' is not
        implemented.
    source : str | constants.Analysis.Settings.SourceGia
        Specifies how the source is defined. Defaults to 'Uniform'.
    number_of_pixels : int
        The number of pixels across the width of the image. Not used when show_as is set to 'SpotDiagram'. Defaults to
        100.
    row_column_number : int | str
        An integer defining the row or column number used when show_as is either 'CrossX' or 'CrossY'. Also accepts
        'Center'. Defaults to 'Center'.
    na : float
        The numerical aperture cut-off. Ignored if set to 0. Defaults to 0.
    total_watts : float
        The total power in watts. Defaults to 1.
    parity : str | constants.Analysis.Settings.Parity
        Defines how the object would appear in object space. Defaults to 'Even'.
    reference : str | constants.Analysis.Settings.ReferenceGia
        Defines the reference coordinate for the center of the plot. Defaults to 'ChiefRay'.
    use_symbols : bool
        Defines whether the plot uses symbols. Only used when show-as is set to 'SpotDiagram'. Defaults to False.
    use_polarization : bool
        Defines whether polarization is considered. Defaults to False.
    remove_vignetting_factors : bool
        Defines whether vignetting factors are automatically removed. Defaults to True.
    scatter_rays : bool
        Defines whether rays are scattered by surfaces with scattering properties. Defaults to False.
    delete_vignetted : False
        Defines whether rays vignetted by any surface are drawn. Defaults to False.
    use_pixel_interpolation : bool
        Defines whether pixel interpolation is used. This setting has no effect when show-as is set to 'SpotDiagram'.
        Defaults to False.
    save_as_bim_file : str
        Defines the filename used to save the output image as BIM file. This file will be stored in the OpticStudio
        images folder. The filename should include the .BIM extension. If set to '', the image will not be saved. Not
        used when show_as is set to 'SpotDiagram.  Defaults to ''.
    """

    wavelength: WavelengthNumber = Field(default="All", description="Wavelength number or 'All'")
    field: int = Field(default=1, gt=0, description="Field number")
    surface: Literal["Image"] | Annotated[int, Field(gt=0)] = Field(default="Image", description="Surface")
    field_size: float = Field(default=0, description="Field size")
    image_size: float = Field(default=50, description="Field size")
    file: str = Field(default="LETTERF.IMA", description="Image file used for the analysis")
    rotation: float = Field(default=0, description="Rotation angle in degrees")
    rays_x_1000: int = Field(default=10, gt=1, description="Approximate number of traced rays (x 1000)")
    show_as: ZOSAPIConstant("Analysis.GiaShowAsTypes") = Field(default="Surface", description="Show As")
    source: ZOSAPIConstant("Analysis.Settings.SourceGia") = Field(default="Uniform", description="Source")
    number_of_pixels: int = Field(default=100, ge=1, description="Number of pixels")
    row_column_number: Literal["Center"] | Annotated[int, Field(gt=0)] = Field(
        default="Center", description="Row or column number"
    )
    na: float = Field(default=0, description="Numerical aperture cut-off")
    total_watts: float = Field(default=1, gt=0, description="Total watts")
    parity: ZOSAPIConstant("Analysis.Settings.Parity") = Field(default="Even", description="Parity")
    reference: ZOSAPIConstant("Analysis.Settings.ReferenceGia") = Field(default="ChiefRay", description="Chief Ray")
    use_symbols: bool = Field(default=False, description="Use symbols")
    use_polarization: bool = Field(default=False, description="Use polarization")
    remove_vignetting_factors: bool = Field(default=True, description="Remove vignetting factors")
    scatter_rays: bool = Field(default=False, description="Scatter rays")
    delete_vignetted: bool = Field(default=False, description="Delete vignetted")
    use_pixel_interpolation: bool = Field(default=False, description="Use pixel interpolation")
    save_as_bim_file: str = Field(default="", description="Filename used to save output as BIM file")


class GeometricImageAnalysis(
    BaseAnalysisWrapper[Union[DataFrame, None], GeometricImageAnalysisSettings], analysis_type="GeometricImageAnalysis"
):
    """Geometric Image Analysis."""

    def __init__(
        self,
        *,
        wavelength: Literal["All"] | int = "All",
        field: int = 1,
        surface: Literal["Image"] | int = "Image",
        field_size: float = 0,
        image_size: float = 50,
        file: str = "LETTERF.IMA",
        rotation: float = 0,
        rays_x_1000: int = 10,
        show_as: constants.Analysis.GiaShowAsTypes | str = "Surface",
        source: constants.Analysis.Settings.SourceGia | str = "Uniform",
        number_of_pixels: int = 100,
        row_column_number: Literal["Center"] | int = "Center",
        na: float = 0,
        total_watts: float = 1,
        parity: constants.Analysis.Settings.Parity | str = "Even",
        reference: constants.Analysis.Settings.ReferenceGia | str = "ChiefRay",
        use_symbols: bool = False,
        use_polarization: bool = False,
        remove_vignetting_factors: bool = True,
        scatter_rays: bool = False,
        delete_vignetted: bool = False,
        use_pixel_interpolation: bool = False,
        save_as_bim_file: str = "",
        settings: GeometricImageAnalysisSettings | None = None,
    ):
        """Create a new Geometric Image Analysis.

        See Also
        --------
        GeometricImageAnalysisSettings : Settings for the Geometric Image Analysis analysis.
        """
        super().__init__(settings_kws=locals())

    def get_data_grid(self, minx=None, miny=None) -> pd.DataFrame | None:
        """Get the data grids from the analysis result.

        Returns
        -------
        pd.DataFrame | None
            The data grids from the analysis result, or None if there are no data grids.
        """
        # Obtain correct origin for datagrid as there is a bug in the API
        # See also https://community.zemax.com/zos-api-12/incorrect-datagrid-miny-for-geometric-image-analysis-5426
        minx = -self.analysis.Settings.ImageSize / 2
        miny = -self.analysis.Settings.ImageSize / 2

        # Get data
        data = [
            zputils.unpack_datagrid(self.analysis.Results.DataGrids[i], minx=minx, miny=miny, cell_origin="bottom_left")
            for i in range(self.analysis.Results.NumberOfDataGrids)
        ]

        return self._process_data_series_or_grid(data)

    def run_analysis(self) -> DataFrame | None:
        """Run the FFT Through Focus MTF analysis."""
        self.analysis.set_wavelength(
            0 if self.settings.wavelength == "All" else self.settings.wavelength
        )  # TODO track with future releases
        self.analysis.set_field(self.settings.field)
        self.analysis.set_surface(self.settings.surface)
        self.analysis.Settings.FieldSize = self.settings.field_size
        self.analysis.Settings.ImageSize = self.settings.image_size
        self.analysis.Settings.File = self.settings.file
        self.analysis.Settings.Rotation = self.settings.rotation
        self.analysis.Settings.RaysX1000 = self.settings.rays_x_1000
        self.analysis.Settings.ShowAs = constants.process_constant(
            constants.Analysis.GiaShowAsTypes, self.settings.show_as
        )
        self.analysis.Settings.Source = constants.process_constant(
            constants.Analysis.Settings.SourceGia, self.settings.source
        )

        if str(self.analysis.Settings.ShowAs) in ("CrossX", "CrossY"):
            if self.settings.row_column_number == "Center":
                self.analysis.Settings.UseColumnRowCenter()
            else:
                self.analysis.Settings.RowColumnNumber = self.settings.row_column_number
        if str(self.analysis.Settings.ShowAs) != "SpotDiagram":
            self.analysis.Settings.NumberOfPixels = self.settings.number_of_pixels
        if str(self.analysis.Settings.ShowAs) == "SpotDiagram":
            self.analysis.Settings.UseSymbols = self.settings.use_symbols

        self.analysis.Settings.NA = self.settings.na
        self.analysis.Settings.TotalWatts = self.settings.total_watts
        self.analysis.Settings.Parity = constants.process_constant(
            constants.Analysis.Settings.Parity, self.settings.parity
        )
        self.analysis.Settings.Reference = constants.process_constant(
            constants.Analysis.Settings.ReferenceGia, self.settings.reference
        )

        self.analysis.Settings.UsePolarization = self.settings.use_polarization
        self.analysis.Settings.RemoveVignettingFactors = self.settings.remove_vignetting_factors
        self.analysis.Settings.ScatterRays = self.settings.scatter_rays
        self.analysis.Settings.DeleteVignetted = self.settings.delete_vignetted
        self.analysis.Settings.UsePixelInterpolation = self.settings.use_pixel_interpolation
        self.analysis.Settings.SaveAsBIMFile = self.settings.save_as_bim_file

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Get results
        return self.get_data_grid()



"""Utilities for working with the Common Language Runtime (CLR)."""

import os
from collections import namedtuple
from datetime import datetime as dt

import clr
from System import Double, Enum, Reflection

DUMMY_DOUBLE = Double(0.0)
DUMMY_ENUM = 0


def clr_get_available_assemblies(*, with_meta=True):
    """Get all the available assemblies from the Common Language Runtime.

    Parameters
    ----------
    with_meta : bool
        Defines if metadata should also be returned

    Returns
    -------
    list
        A list of all available assemblies
    """
    return list(clr.ListAssemblies(with_meta))


def reflect_dll_content(dllfilepath):
    """Reflects the content of a dll.

    Parameters
    ----------
    dllfilepath : str
        The absolute path to the dll, including extension.

    Returns
    -------
    dict
        A dictionary containing the namespaces ('namespaces') and the enumerations ('enums').
    """
    if not os.path.isabs(dllfilepath):
        raise ValueError("dllfilepath should be an absolute path")
    if not dllfilepath.lower().endswith(".dll"):
        raise ValueError("dllfilepath should end with .dll (case is ignored)")

    content = list(Reflection.Assembly.LoadFile(dllfilepath).GetTypes())

    namespaces = sorted({item.Namespace for item in content})
    enums = sorted([item.FullName for item in content if item.IsEnum])
    return {"namespaces": namespaces, "enums": enums}


def system_get_enum_key_from_value(enum, value):
    """Get the key corresponding to a certain value from a System.Enum instance.

    Parameters
    ----------
    enum : System.Enum
        The Enum instance.
    value
        The value for which the key should be returned

    Returns
    -------
    Any
        The corresponding key
    """
    return Enum.GetName(enum, value)


def system_get_enum_names(enum):
    """Get all names from a System.Enum instance.

    Parameters
    ----------
    enum : System.Enum
        An Enum instance.

    Returns
    -------
    list
        A list containing the names of the Enum instance
    """
    return list(Enum.GetNames(enum))


def system_get_enum_values(enum):
    """Get all values from a System.Enum instance.

    Parameters
    ----------
    enum : System.Enum
        An Enum instance.

    Returns
    -------
    list
        A list containing the values of the Enum instance
    """
    return list(Enum.GetValues(enum))


def system_enum_to_namedtuple(enum):
    """Convert a System.Enum into a namedtuple.

    If the Enum contains a member with name `None`, its name is changed to `None_`.

    Parameters
    ----------
    enum : System.Enum
        An Enum instance.

    Returns
    -------
    namedtuple
        A namedtuple containing the enum data.
    """
    label = enum.__name__
    names = system_get_enum_names(enum)
    values = system_get_enum_values(enum)

    # Rename None to NONE
    if "None" in names:
        names[names.index("None")] = "None_"

    return namedtuple(label, names)._make(values)  # noqa


def system_datetime_to_datetime(sdt):
    """Convert a System.DateTime into a datetime.datetime instance.

    Note: As the System.DateTime does not contain info on the timezone, there might be an error if these do not match
    between the supplied DateTime instance and the current computer

    Parameters
    ----------
    sdt : System.DateTime
        An System.DateTime instance

    Returns
    -------
    datetime.datetime
        The pyton datetime instance
    """
    return dt(  # noqa: DTZ001
        year=sdt.Year,
        month=sdt.Month,
        day=sdt.Day,
        hour=sdt.Hour,
        minute=sdt.Minute,
        second=sdt.Second,
        microsecond=sdt.Millisecond * 1000,
    )



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "CoatingStatusType",
    "ConversionOrder",
    "CoordinateConversionResult",
    "CoordinateReturnType",
    "IndexDataType",
    "InterpolationMethod",
    "NodesDataType",
    "PilotRadiusMode",
    "PointCloudFileFormat",
    "PupilApodizationType",
    "QTypes",
    "STARDeformationOption",
    "SubstrateType",
    "SurfaceApertureTypes",
    "SurfaceColumn",
    "SurfaceEdgeDraw",
    "SurfaceScatteringTypes",
    "SurfaceType",
    "TiltDecenterOrderType",
    "TiltDecenterPickupType",
    "TiltType",
    "XYSampling",
)

class CoatingStatusType:
    Fixed: CoatingStatusType = None
    Variable: CoatingStatusType = None
    Pickup: CoatingStatusType = None

class ConversionOrder:
    Forward: ConversionOrder = None
    Reverse: ConversionOrder = None

class CoordinateConversionResult:
    Success: CoordinateConversionResult = None
    Error_MultiConfig: CoordinateConversionResult = None
    Error_TiltDecenter: CoordinateConversionResult = None
    Error_IgnoredSurface: CoordinateConversionResult = None
    Error_CoordianteBreak: CoordinateConversionResult = None
    Error_InvalidRange: CoordinateConversionResult = None

class CoordinateReturnType:
    None_: CoordinateReturnType = None
    OrientationOnly: CoordinateReturnType = None
    OrientationXY: CoordinateReturnType = None
    OrientationXYZ: CoordinateReturnType = None

class IndexDataType:
    None_: IndexDataType = None
    PhysicsBasedIndex: IndexDataType = None
    DirectRefractiveIndex: IndexDataType = None
    Unknown: IndexDataType = None

class InterpolationMethod:
    BicubicSpline: InterpolationMethod = None
    Linear: InterpolationMethod = None

class NodesDataType:
    SurfaceDeformationNoRBM: NodesDataType = None
    RefractiveIndex: NodesDataType = None
    TemperatureAndRefractiveIndex: NodesDataType = None
    SurfaceDeformation: NodesDataType = None

class PilotRadiusMode:
    BestFit: PilotRadiusMode = None
    Shorter: PilotRadiusMode = None
    Longer: PilotRadiusMode = None
    X: PilotRadiusMode = None
    Y: PilotRadiusMode = None
    Plane: PilotRadiusMode = None
    User: PilotRadiusMode = None

class PointCloudFileFormat:
    ASCII: PointCloudFileFormat = None
    Binary: PointCloudFileFormat = None
    CompressedBinary: PointCloudFileFormat = None

class PupilApodizationType:
    None_: PupilApodizationType = None
    Gaussian: PupilApodizationType = None
    Tangential: PupilApodizationType = None

class QTypes:
    Qbfs: QTypes = None
    Qcon: QTypes = None

class STARDeformationOption:
    DeformationWithRBMs: STARDeformationOption = None
    DeformationWithoutRBMs: STARDeformationOption = None
    OnlyRBMs: STARDeformationOption = None
    NoDeformation: STARDeformationOption = None

class SubstrateType:
    None_: SubstrateType = None
    Flat: SubstrateType = None
    Curved: SubstrateType = None

class SurfaceApertureTypes:
    None_: SurfaceApertureTypes = None
    CircularAperture: SurfaceApertureTypes = None
    CircularObscuration: SurfaceApertureTypes = None
    Spider: SurfaceApertureTypes = None
    RectangularAperture: SurfaceApertureTypes = None
    RectangularObscuration: SurfaceApertureTypes = None
    EllipticalAperture: SurfaceApertureTypes = None
    EllipticalObscuration: SurfaceApertureTypes = None
    UserAperture: SurfaceApertureTypes = None
    UserObscuration: SurfaceApertureTypes = None
    FloatingAperture: SurfaceApertureTypes = None

class SurfaceColumn:
    Comment: SurfaceColumn = None
    Radius: SurfaceColumn = None
    Thickness: SurfaceColumn = None
    Material: SurfaceColumn = None
    Coating: SurfaceColumn = None
    SemiDiameter: SurfaceColumn = None
    ChipZone: SurfaceColumn = None
    MechanicalSemiDiameter: SurfaceColumn = None
    Conic: SurfaceColumn = None
    TCE: SurfaceColumn = None
    Par0: SurfaceColumn = None
    Par1: SurfaceColumn = None
    Par2: SurfaceColumn = None
    Par3: SurfaceColumn = None
    Par4: SurfaceColumn = None
    Par5: SurfaceColumn = None
    Par6: SurfaceColumn = None
    Par7: SurfaceColumn = None
    Par8: SurfaceColumn = None
    Par9: SurfaceColumn = None
    Par10: SurfaceColumn = None
    Par11: SurfaceColumn = None
    Par12: SurfaceColumn = None
    Par13: SurfaceColumn = None
    Par14: SurfaceColumn = None
    Par15: SurfaceColumn = None
    Par16: SurfaceColumn = None
    Par17: SurfaceColumn = None
    Par18: SurfaceColumn = None
    Par19: SurfaceColumn = None
    Par20: SurfaceColumn = None
    Par21: SurfaceColumn = None
    Par22: SurfaceColumn = None
    Par23: SurfaceColumn = None
    Par24: SurfaceColumn = None
    Par25: SurfaceColumn = None
    Par26: SurfaceColumn = None
    Par27: SurfaceColumn = None
    Par28: SurfaceColumn = None
    Par29: SurfaceColumn = None
    Par30: SurfaceColumn = None
    Par31: SurfaceColumn = None
    Par32: SurfaceColumn = None
    Par33: SurfaceColumn = None
    Par34: SurfaceColumn = None
    Par35: SurfaceColumn = None
    Par36: SurfaceColumn = None
    Par37: SurfaceColumn = None
    Par38: SurfaceColumn = None
    Par39: SurfaceColumn = None
    Par40: SurfaceColumn = None
    Par41: SurfaceColumn = None
    Par42: SurfaceColumn = None
    Par43: SurfaceColumn = None
    Par44: SurfaceColumn = None
    Par45: SurfaceColumn = None
    Par46: SurfaceColumn = None
    Par47: SurfaceColumn = None
    Par48: SurfaceColumn = None
    Par49: SurfaceColumn = None
    Par50: SurfaceColumn = None
    Par51: SurfaceColumn = None
    Par52: SurfaceColumn = None
    Par53: SurfaceColumn = None
    Par54: SurfaceColumn = None
    Par55: SurfaceColumn = None
    Par56: SurfaceColumn = None
    Par57: SurfaceColumn = None
    Par58: SurfaceColumn = None
    Par59: SurfaceColumn = None
    Par60: SurfaceColumn = None
    Par61: SurfaceColumn = None
    Par62: SurfaceColumn = None
    Par63: SurfaceColumn = None
    Par64: SurfaceColumn = None
    Par65: SurfaceColumn = None
    Par66: SurfaceColumn = None
    Par67: SurfaceColumn = None
    Par68: SurfaceColumn = None
    Par69: SurfaceColumn = None
    Par70: SurfaceColumn = None
    Par71: SurfaceColumn = None
    Par72: SurfaceColumn = None
    Par73: SurfaceColumn = None
    Par74: SurfaceColumn = None
    Par75: SurfaceColumn = None
    Par76: SurfaceColumn = None
    Par77: SurfaceColumn = None
    Par78: SurfaceColumn = None
    Par79: SurfaceColumn = None
    Par80: SurfaceColumn = None
    Par81: SurfaceColumn = None
    Par82: SurfaceColumn = None
    Par83: SurfaceColumn = None
    Par84: SurfaceColumn = None
    Par85: SurfaceColumn = None
    Par86: SurfaceColumn = None
    Par87: SurfaceColumn = None
    Par88: SurfaceColumn = None
    Par89: SurfaceColumn = None
    Par90: SurfaceColumn = None
    Par91: SurfaceColumn = None
    Par92: SurfaceColumn = None
    Par93: SurfaceColumn = None
    Par94: SurfaceColumn = None
    Par95: SurfaceColumn = None
    Par96: SurfaceColumn = None
    Par97: SurfaceColumn = None
    Par98: SurfaceColumn = None
    Par99: SurfaceColumn = None
    Par100: SurfaceColumn = None
    Par101: SurfaceColumn = None
    Par102: SurfaceColumn = None
    Par103: SurfaceColumn = None
    Par104: SurfaceColumn = None
    Par105: SurfaceColumn = None
    Par106: SurfaceColumn = None
    Par107: SurfaceColumn = None
    Par108: SurfaceColumn = None
    Par109: SurfaceColumn = None
    Par110: SurfaceColumn = None
    Par111: SurfaceColumn = None
    Par112: SurfaceColumn = None
    Par113: SurfaceColumn = None
    Par114: SurfaceColumn = None
    Par115: SurfaceColumn = None
    Par116: SurfaceColumn = None
    Par117: SurfaceColumn = None
    Par118: SurfaceColumn = None
    Par119: SurfaceColumn = None
    Par120: SurfaceColumn = None
    Par121: SurfaceColumn = None
    Par122: SurfaceColumn = None
    Par123: SurfaceColumn = None
    Par124: SurfaceColumn = None
    Par125: SurfaceColumn = None
    Par126: SurfaceColumn = None
    Par127: SurfaceColumn = None
    Par128: SurfaceColumn = None
    Par129: SurfaceColumn = None
    Par130: SurfaceColumn = None
    Par131: SurfaceColumn = None
    Par132: SurfaceColumn = None
    Par133: SurfaceColumn = None
    Par134: SurfaceColumn = None
    Par135: SurfaceColumn = None
    Par136: SurfaceColumn = None
    Par137: SurfaceColumn = None
    Par138: SurfaceColumn = None
    Par139: SurfaceColumn = None
    Par140: SurfaceColumn = None
    Par141: SurfaceColumn = None
    Par142: SurfaceColumn = None
    Par143: SurfaceColumn = None
    Par144: SurfaceColumn = None
    Par145: SurfaceColumn = None
    Par146: SurfaceColumn = None
    Par147: SurfaceColumn = None
    Par148: SurfaceColumn = None
    Par149: SurfaceColumn = None
    Par150: SurfaceColumn = None
    Par151: SurfaceColumn = None
    Par152: SurfaceColumn = None
    Par153: SurfaceColumn = None
    Par154: SurfaceColumn = None
    Par155: SurfaceColumn = None
    Par156: SurfaceColumn = None
    Par157: SurfaceColumn = None
    Par158: SurfaceColumn = None
    Par159: SurfaceColumn = None
    Par160: SurfaceColumn = None
    Par161: SurfaceColumn = None
    Par162: SurfaceColumn = None
    Par163: SurfaceColumn = None
    Par164: SurfaceColumn = None
    Par165: SurfaceColumn = None
    Par166: SurfaceColumn = None
    Par167: SurfaceColumn = None
    Par168: SurfaceColumn = None
    Par169: SurfaceColumn = None
    Par170: SurfaceColumn = None
    Par171: SurfaceColumn = None
    Par172: SurfaceColumn = None
    Par173: SurfaceColumn = None
    Par174: SurfaceColumn = None
    Par175: SurfaceColumn = None
    Par176: SurfaceColumn = None
    Par177: SurfaceColumn = None
    Par178: SurfaceColumn = None
    Par179: SurfaceColumn = None
    Par180: SurfaceColumn = None
    Par181: SurfaceColumn = None
    Par182: SurfaceColumn = None
    Par183: SurfaceColumn = None
    Par184: SurfaceColumn = None
    Par185: SurfaceColumn = None
    Par186: SurfaceColumn = None
    Par187: SurfaceColumn = None
    Par188: SurfaceColumn = None
    Par189: SurfaceColumn = None
    Par190: SurfaceColumn = None
    Par191: SurfaceColumn = None
    Par192: SurfaceColumn = None
    Par193: SurfaceColumn = None
    Par194: SurfaceColumn = None
    Par195: SurfaceColumn = None
    Par196: SurfaceColumn = None
    Par197: SurfaceColumn = None
    Par198: SurfaceColumn = None
    Par199: SurfaceColumn = None
    Par200: SurfaceColumn = None
    Par201: SurfaceColumn = None
    Par202: SurfaceColumn = None
    Par203: SurfaceColumn = None
    Par204: SurfaceColumn = None
    Par205: SurfaceColumn = None
    Par206: SurfaceColumn = None
    Par207: SurfaceColumn = None
    Par208: SurfaceColumn = None
    Par209: SurfaceColumn = None
    Par210: SurfaceColumn = None
    Par211: SurfaceColumn = None
    Par212: SurfaceColumn = None
    Par213: SurfaceColumn = None
    Par214: SurfaceColumn = None
    Par215: SurfaceColumn = None
    Par216: SurfaceColumn = None
    Par217: SurfaceColumn = None
    Par218: SurfaceColumn = None
    Par219: SurfaceColumn = None
    Par220: SurfaceColumn = None
    Par221: SurfaceColumn = None
    Par222: SurfaceColumn = None
    Par223: SurfaceColumn = None
    Par224: SurfaceColumn = None
    Par225: SurfaceColumn = None
    Par226: SurfaceColumn = None
    Par227: SurfaceColumn = None
    Par228: SurfaceColumn = None
    Par229: SurfaceColumn = None
    Par230: SurfaceColumn = None
    Par231: SurfaceColumn = None
    Par232: SurfaceColumn = None
    Par233: SurfaceColumn = None
    Par234: SurfaceColumn = None
    Par235: SurfaceColumn = None
    Par236: SurfaceColumn = None
    Par237: SurfaceColumn = None
    Par238: SurfaceColumn = None
    Par239: SurfaceColumn = None
    Par240: SurfaceColumn = None
    Par241: SurfaceColumn = None
    Par242: SurfaceColumn = None
    Par243: SurfaceColumn = None
    Par244: SurfaceColumn = None
    Par245: SurfaceColumn = None
    Par246: SurfaceColumn = None
    Par247: SurfaceColumn = None
    Par248: SurfaceColumn = None
    Par249: SurfaceColumn = None
    Par250: SurfaceColumn = None
    Par251: SurfaceColumn = None
    Par252: SurfaceColumn = None
    Par253: SurfaceColumn = None
    Par254: SurfaceColumn = None

class SurfaceEdgeDraw:
    Squared: SurfaceEdgeDraw = None
    Tapered: SurfaceEdgeDraw = None
    Flat: SurfaceEdgeDraw = None

class SurfaceScatteringTypes:
    None_: SurfaceScatteringTypes = None
    Lambertian: SurfaceScatteringTypes = None
    Gaussian: SurfaceScatteringTypes = None
    ABg: SurfaceScatteringTypes = None
    ABgFile: SurfaceScatteringTypes = None
    BSDF: SurfaceScatteringTypes = None
    User: SurfaceScatteringTypes = None
    ISScatterCatalog: SurfaceScatteringTypes = None

class SurfaceType:
    ABCD: SurfaceType = None
    AlternateEven: SurfaceType = None
    AlternateOdd: SurfaceType = None
    AnnularZernikeSag: SurfaceType = None
    Atmospheric: SurfaceType = None
    Biconic: SurfaceType = None
    BiconicZernike: SurfaceType = None
    Binary1: SurfaceType = None
    Binary2: SurfaceType = None
    Binary3: SurfaceType = None
    Binary4: SurfaceType = None
    BirefringentIn: SurfaceType = None
    BirefringentOut: SurfaceType = None
    BlackBoxLens: SurfaceType = None
    ChebyShv: SurfaceType = None
    Conjugate: SurfaceType = None
    CoordinateBreak: SurfaceType = None
    CubicSpline: SurfaceType = None
    CylinderFrensel: SurfaceType = None
    CylinderFresnel: SurfaceType = None
    Data: SurfaceType = None
    DiffractionGrating: SurfaceType = None
    EllipticalGrating1: SurfaceType = None
    EllipticalGrating2: SurfaceType = None
    EvenAspheric: SurfaceType = None
    ExtendedToroidalGrating: SurfaceType = None
    ExtendedAsphere: SurfaceType = None
    ExtendedCubicSpline: SurfaceType = None
    ExtendedFresnel: SurfaceType = None
    ExtendedOddAsphere: SurfaceType = None
    ExtendedPolynomial: SurfaceType = None
    Fresnel: SurfaceType = None
    GeneralizedFresnel: SurfaceType = None
    Gradient1: SurfaceType = None
    Gradient2: SurfaceType = None
    Gradient3: SurfaceType = None
    Gradient4: SurfaceType = None
    Gradient5: SurfaceType = None
    Gradient6: SurfaceType = None
    Gradient7: SurfaceType = None
    Gradient9: SurfaceType = None
    Gradient10: SurfaceType = None
    Gradient12: SurfaceType = None
    Gradium: SurfaceType = None
    GridGradient: SurfaceType = None
    GridPhase: SurfaceType = None
    GridSag: SurfaceType = None
    Hologram1: SurfaceType = None
    Hologram2: SurfaceType = None
    Irregular: SurfaceType = None
    JonesMatrix: SurfaceType = None
    NonSequential: SurfaceType = None
    OddAsphere: SurfaceType = None
    OddCosine: SurfaceType = None
    OffAxisConicFreeform: SurfaceType = None
    OpticallyFabricatedHologram: SurfaceType = None
    Paraxial: SurfaceType = None
    ParaxialXY: SurfaceType = None
    Periodic: SurfaceType = None
    Polynomial: SurfaceType = None
    QTypeAsphere: SurfaceType = None
    QTypeFreeform: SurfaceType = None
    RadialGrating: SurfaceType = None
    RadialNurbs: SurfaceType = None
    RetroReflect: SurfaceType = None
    SlideSurface: SurfaceType = None
    Standard: SurfaceType = None
    Superconic: SurfaceType = None
    Tilted: SurfaceType = None
    Toroidal: SurfaceType = None
    ToroidalGrat: SurfaceType = None
    ToroidalHologram: SurfaceType = None
    ToroidalNurbs: SurfaceType = None
    UserDefined: SurfaceType = None
    VariableLineSpaceGrating: SurfaceType = None
    ZernikeAnnularPhase: SurfaceType = None
    ZernikeFringePhase: SurfaceType = None
    ZernikeFringeSag: SurfaceType = None
    ZernikeStandardPhase: SurfaceType = None
    ZernikeStandardSag: SurfaceType = None
    ZonePlate: SurfaceType = None
    Freeform: SurfaceType = None

class TiltDecenterOrderType:
    Decenter_Tilt: TiltDecenterOrderType = None
    Tilt_Decenter: TiltDecenterOrderType = None

class TiltDecenterPickupType:
    Explicit: TiltDecenterPickupType = None
    PickupSurface: TiltDecenterPickupType = None
    ReverseSurface: TiltDecenterPickupType = None

class TiltType:
    XTilt: TiltType = None
    YTilt: TiltType = None

class XYSampling:
    S32: XYSampling = None
    S64: XYSampling = None
    S128: XYSampling = None
    S256: XYSampling = None
    S512: XYSampling = None
    S1024: XYSampling = None
    S2048: XYSampling = None
    S4096: XYSampling = None
    S8192: XYSampling = None
    S16384: XYSampling = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = "PathAnalysisSortType"

class PathAnalysisSortType:
    TotalEndingFlux: PathAnalysisSortType = None
    NumberOfObjectsStruck: PathAnalysisSortType = None
    NumberOfBranches: PathAnalysisSortType = None
    LastObjectStruck: PathAnalysisSortType = None
    NumberOfUniqueObjectsStruct: PathAnalysisSortType = None
    AverageOpticalPathLength: PathAnalysisSortType = None
    NumberOfDiffractionEvents: PathAnalysisSortType = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import SampleSizes
from zospy.api._ZOSAPI.Analysis.Settings import IAS_

__all__ = (
    "IAS_PhysicalOpticsPropagation",
    "POPBeamTypes",
    "POPDataTypes",
    "POPFiberPositions",
    "POPFiberTypes",
    "POPProjectionTypes",
    "POPScaleTypes",
    "POPShowAsTypes",
    "POPZoomTypes",
)

class IAS_PhysicalOpticsPropagation(IAS_):
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def StartSurface(self) -> IAS_Surface: ...
    @property
    def EndSurface(self) -> IAS_Surface: ...
    @property
    def SurfaceToBeam(self) -> float: ...
    @SurfaceToBeam.setter
    def SurfaceToBeam(self, value: float) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def SeparateXY(self) -> bool: ...
    @SeparateXY.setter
    def SeparateXY(self, value: bool) -> None: ...
    @property
    def UseDiskStorage(self) -> bool: ...
    @UseDiskStorage.setter
    def UseDiskStorage(self, value: bool) -> None: ...
    @property
    def BeamType(self) -> POPBeamTypes: ...
    @BeamType.setter
    def BeamType(self, value: POPBeamTypes) -> None: ...
    @property
    def XSampling(self) -> SampleSizes: ...
    @XSampling.setter
    def XSampling(self, value: SampleSizes) -> None: ...
    @property
    def YSampling(self) -> SampleSizes: ...
    @YSampling.setter
    def YSampling(self, value: SampleSizes) -> None: ...
    @property
    def XWidth(self) -> float: ...
    @XWidth.setter
    def XWidth(self, value: float) -> None: ...
    @property
    def YWidth(self) -> float: ...
    @YWidth.setter
    def YWidth(self, value: float) -> None: ...
    @property
    def UseTotalPower(self) -> bool: ...
    @UseTotalPower.setter
    def UseTotalPower(self, value: bool) -> None: ...
    @property
    def TotalPower(self) -> float: ...
    @TotalPower.setter
    def TotalPower(self, value: float) -> None: ...
    @property
    def UsePeakIrradiance(self) -> bool: ...
    @UsePeakIrradiance.setter
    def UsePeakIrradiance(self, value: bool) -> None: ...
    @property
    def PeakIrradiance(self) -> float: ...
    @PeakIrradiance.setter
    def PeakIrradiance(self, value: float) -> None: ...
    @property
    def NumberOfBeamTypeFiles(self) -> int: ...
    @property
    def BeamTypeFilename(self) -> str: ...
    @BeamTypeFilename.setter
    def BeamTypeFilename(self, value: str) -> None: ...
    @property
    def NumberOfParameters(self) -> int: ...
    @property
    def ShowAs(self) -> POPShowAsTypes: ...
    @ShowAs.setter
    def ShowAs(self, value: POPShowAsTypes) -> None: ...
    @property
    def DataType(self) -> POPDataTypes: ...
    @DataType.setter
    def DataType(self, value: POPDataTypes) -> None: ...
    @property
    def Project(self) -> POPProjectionTypes: ...
    @Project.setter
    def Project(self, value: POPProjectionTypes) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def PlotScale(self) -> float: ...
    @PlotScale.setter
    def PlotScale(self, value: float) -> None: ...
    @property
    def LogScale(self) -> POPScaleTypes: ...
    @LogScale.setter
    def LogScale(self, value: POPScaleTypes) -> None: ...
    @property
    def ZoomIn(self) -> POPZoomTypes: ...
    @ZoomIn.setter
    def ZoomIn(self, value: POPZoomTypes) -> None: ...
    @property
    def RowOrColumn(self) -> int: ...
    @RowOrColumn.setter
    def RowOrColumn(self, value: int) -> None: ...
    @property
    def ZeroPhaseLevel(self) -> float: ...
    @ZeroPhaseLevel.setter
    def ZeroPhaseLevel(self, value: float) -> None: ...
    @property
    def SaveOutputBeam(self) -> bool: ...
    @SaveOutputBeam.setter
    def SaveOutputBeam(self, value: bool) -> None: ...
    @property
    def OutputBeamFile(self) -> str: ...
    @OutputBeamFile.setter
    def OutputBeamFile(self, value: str) -> None: ...
    @property
    def SaveBeamAtAllSurfaces(self) -> bool: ...
    @SaveBeamAtAllSurfaces.setter
    def SaveBeamAtAllSurfaces(self, value: bool) -> None: ...
    @property
    def ComputeFiberCouplingIntegral(self) -> bool: ...
    @ComputeFiberCouplingIntegral.setter
    def ComputeFiberCouplingIntegral(self, value: bool) -> None: ...
    @property
    def IgnoreFiberPolarization(self) -> bool: ...
    @IgnoreFiberPolarization.setter
    def IgnoreFiberPolarization(self, value: bool) -> None: ...
    @property
    def FiberType(self) -> POPFiberTypes: ...
    @FiberType.setter
    def FiberType(self, value: POPFiberTypes) -> None: ...
    @property
    def FiberPosition(self) -> POPFiberPositions: ...
    @FiberPosition.setter
    def FiberPosition(self, value: POPFiberPositions) -> None: ...
    @property
    def TiltAboutX(self) -> float: ...
    @TiltAboutX.setter
    def TiltAboutX(self, value: float) -> None: ...
    @property
    def TiltAboutY(self) -> float: ...
    @TiltAboutY.setter
    def TiltAboutY(self, value: float) -> None: ...
    @property
    def NumberOfFiberTypeFiles(self) -> int: ...
    @property
    def FiberTypeFilename(self) -> str: ...
    @FiberTypeFilename.setter
    def FiberTypeFilename(self, value: str) -> None: ...
    @property
    def NumberOfFiberParameters(self) -> int: ...
    def AutoCalculateBeamSampling(self) -> None: ...
    def GetBeamTypeFileName(self, fileIndex: int) -> str: ...
    def GetBeamTypeFiles(self) -> list[str]: ...
    def GetFiberParameterName(self, paramIdx: int) -> str: ...
    def GetFiberParameterValue(self, paramIdx: int) -> float: ...
    def GetFiberTypeFileName(self, fileIndex: int) -> str: ...
    def GetFiberTypeFiles(self) -> list[str]: ...
    def GetParameterName(self, paramIdx: int) -> str: ...
    def GetParameterValue(self, paramIdx: int) -> float: ...
    def ReadPowerFromFile(self) -> None: ...
    def SetFiberParameterValue(self, paramIdx: int, Value: float) -> None: ...
    def SetParameterValue(self, paramIdx: int, Value: float) -> None: ...

class POPBeamTypes:
    GaussianWaist = 0
    GaussianAngle = 1
    GaussianSizeAngle = 2
    TopHat = 3
    File = 4
    DLL = 5
    Multimode = 6
    AstigmaticGaussian = 7

class POPDataTypes:
    Irradiance = 0
    EXIrradiance = 1
    EYIrradiance = 2
    Phase = 3
    EXPhase = 4
    EYPhase = 5
    TransferMagnitude = 6
    TransferPhase = 7

class POPFiberPositions:
    ChiefRay = 0
    SurfaceVertex = 1

class POPFiberTypes:
    GaussianWaist = 0
    GaussianAngle = 1
    GaussianSizeAngle = 2
    TopHat = 3
    File = 4
    DLL = 5
    AstigmaticGaussian = 6

class POPProjectionTypes:
    AlongBeam = 0
    AlongNormal = 1
    AlongLocalZ = 2

class POPScaleTypes:
    Linear = 0
    Log_Minus_5 = 1
    Log_Minus_10 = 2
    Log_Minus_15 = 3

class POPShowAsTypes:
    Surface = 0
    Contour = 1
    GrayScale = 2
    InverseGrayScale = 3
    FalseColor = 4
    InverseFalseColor = 5
    CrossX = 6
    CrossY = 7
    Encircled = 8
    Ensquared = 9
    EnslittedX = 10
    EnslittedY = 11

class POPZoomTypes:
    NoZoom = 0
    X2 = 1
    X4 = 2
    X8 = 3
    X16 = 4



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import IColorTranslator, IMessage
from zospy.api._ZOSAPI.Tools.RayTrace import RayStatus

__all__ = (
    "IAR_",
    "IAR_CriticalRayData",
    "IAR_CriticalRayInfo",
    "IAR_DataGrid",
    "IAR_DataGridRgb",
    "IAR_DataGridVector",
    "IAR_DataScatterPoints",
    "IAR_DataScatterPointsRgb",
    "IAR_DataSeries",
    "IAR_DataSeriesRgb",
    "IAR_HeaderData",
    "IAR_MetaData",
    "IAR_NSCSingleRayTraceData",
    "IAR_PathAnalysisData",
    "IAR_PathAnalysisEntry",
    "IAR_RayData",
    "IAR_RayInfo",
    "IAR_Rgb",
    "IAR_ScatterPoint",
    "IAR_ScatterPointRgb",
    "IAR_SpotDataResult",
    "IAR_SpotDataResultMatrix",
    "IAR_Vector2D",
    "IAR_XYZ",
)

class IAR_:
    @property
    def MetaData(self) -> IAR_MetaData: ...
    @property
    def HeaderData(self) -> IAR_HeaderData: ...
    @property
    def NumberOfDataGrids(self) -> int: ...
    @property
    def DataGrids(self) -> list[IAR_DataGrid]: ...
    @property
    def NumberOfDataGridsRgb(self) -> int: ...
    @property
    def DataGridsRgb(self) -> list[IAR_DataGridRgb]: ...
    @property
    def NumberOfDataSeries(self) -> int: ...
    @property
    def DataSeries(self) -> list[IAR_DataSeries]: ...
    @property
    def NumberOfDataSeriesRgb(self) -> int: ...
    @property
    def DataSeriesRgb(self) -> list[IAR_DataSeriesRgb]: ...
    @property
    def NumberOfDataScatterPoints(self) -> int: ...
    @property
    def DataScatterPoints(self) -> list[IAR_DataScatterPoints]: ...
    @property
    def NumberOfDataScatterPointsRgb(self) -> int: ...
    @property
    def DataScatterPointsRgb(self) -> list[IAR_DataScatterPointsRgb]: ...
    @property
    def NumberOfRayData(self) -> int: ...
    @property
    def RayData(self) -> list[IAR_RayData]: ...
    @property
    def CriticalRayData(self) -> IAR_CriticalRayData: ...
    @property
    def PathAnalysisData(self) -> IAR_PathAnalysisData: ...
    @property
    def SpotData(self) -> IAR_SpotDataResultMatrix: ...
    @property
    def NSCSingleRayTraceData(self) -> IAR_NSCSingleRayTraceData: ...
    @property
    def NumberOfMessages(self) -> int: ...
    @property
    def Messages(self) -> list[IMessage]: ...
    def GetDataGrid(self, index: int) -> IAR_DataGrid: ...
    def GetDataGridRgb(self, index: int) -> IAR_DataGridRgb: ...
    def GetDataScatterPoint(self, index: int) -> IAR_DataScatterPoints: ...
    def GetDataScatterPointRgb(self, index: int) -> IAR_DataScatterPointsRgb: ...
    def GetDataSeries(self, index: int) -> IAR_DataSeries: ...
    def GetDataSeriesRgb(self, index: int) -> IAR_DataSeriesRgb: ...
    def GetMessageAt(self, index: int) -> IMessage: ...
    def GetRayData(self, index: int) -> IAR_RayData: ...
    def GetTextFile(self, Filename: str) -> bool: ...

class IAR_CriticalRayData:
    @property
    def NumRays(self) -> int: ...
    @property
    def HeaderLabels(self) -> list[str]: ...
    @property
    def Rays(self) -> list[IAR_CriticalRayInfo]: ...
    def GetRay(self, idx: int) -> IAR_CriticalRayInfo: ...

class IAR_CriticalRayInfo:
    @property
    def FieldPoint(self) -> int: ...
    @property
    def RayType(self) -> CriticalRayType: ...
    @property
    def Pass(self) -> bool: ...
    @property
    def TerminationObject(self) -> int: ...
    @property
    def Wavelength(self) -> float: ...
    @property
    def XIn(self) -> float: ...
    @property
    def YIn(self) -> float: ...
    @property
    def ZIn(self) -> float: ...
    @property
    def LIn(self) -> float: ...
    @property
    def MIn(self) -> float: ...
    @property
    def NIn(self) -> float: ...
    @property
    def XTarget(self) -> float: ...
    @property
    def YTarget(self) -> float: ...
    @property
    def ZTarget(self) -> float: ...
    @property
    def LTarget(self) -> float: ...
    @property
    def MTarget(self) -> float: ...
    @property
    def NTarget(self) -> float: ...
    @property
    def XActual(self) -> float: ...
    @property
    def YActual(self) -> float: ...
    @property
    def ZActual(self) -> float: ...
    @property
    def LActual(self) -> float: ...
    @property
    def MActual(self) -> float: ...
    @property
    def NActual(self) -> float: ...

class IAR_DataGrid:
    @property
    def Description(self) -> str: ...
    @property
    def XLabel(self) -> str: ...
    @property
    def YLabel(self) -> str: ...
    @property
    def ValueLabel(self) -> str: ...
    @property
    def Nx(self) -> int: ...
    @property
    def Ny(self) -> int: ...
    @property
    def Dx(self) -> float: ...
    @property
    def Dy(self) -> float: ...
    @property
    def MinX(self) -> float: ...
    @property
    def MinY(self) -> float: ...
    @property
    def Values(self) -> list[list[float]]: ...
    @property
    def ValueData(self) -> IMatrixData: ...
    def ConvertToRGB(self, translator: IColorTranslator) -> IAR_DataGridRgb: ...
    def X(self, rowX: int) -> float: ...
    def XYZ(self, rowX: int, colY: int) -> IAR_XYZ: ...
    def Y(self, colY: int) -> float: ...
    def Z(self, rowX: int, colY: int) -> float: ...

class IAR_DataGridRgb:
    @property
    def Description(self) -> str: ...
    @property
    def XLabel(self) -> str: ...
    @property
    def YLabel(self) -> str: ...
    @property
    def ValueLabel(self) -> str: ...
    @property
    def Nx(self) -> int: ...
    @property
    def Ny(self) -> int: ...
    @property
    def Dx(self) -> float: ...
    @property
    def Dy(self) -> float: ...
    @property
    def MinX(self) -> float: ...
    @property
    def MinY(self) -> float: ...
    @property
    def Values(self) -> list[list[IAR_Rgb]]: ...
    def FillValues(self, fullSize: int) -> tuple[list[float], list[float], list[float]]: ...
    def GetValue(self, x: int, y: int) -> IAR_Rgb: ...

class IAR_DataGridVector:
    @property
    def Description(self) -> str: ...
    @property
    def XLabel(self) -> str: ...
    @property
    def YLabel(self) -> str: ...
    @property
    def ValueLabel(self) -> str: ...
    @property
    def Nx(self) -> int: ...
    @property
    def Ny(self) -> int: ...
    @property
    def Dx(self) -> float: ...
    @property
    def Dy(self) -> float: ...
    @property
    def MinX(self) -> float: ...
    @property
    def MinY(self) -> float: ...
    @property
    def DeltaXValues(self) -> list[list[float]]: ...
    @property
    def DeltaYValues(self) -> list[list[float]]: ...
    @property
    def DeltaXValueData(self) -> IMatrixData: ...
    @property
    def DeltaYValueData(self) -> IMatrixData: ...
    def DeltaX(self, rowX: int, colY: int) -> float: ...
    def DeltaY(self, rowX: int, colY: int) -> float: ...
    def X(self, rowX: int) -> float: ...
    def Y(self, colY: int) -> float: ...

class IAR_DataScatterPoints:
    @property
    def Description(self) -> str: ...
    @property
    def NumPoints(self) -> int: ...
    @property
    def Points(self) -> list[IAR_ScatterPoint]: ...
    @property
    def XLabel(self) -> str: ...
    @property
    def YLabel(self) -> str: ...
    @property
    def ValueLabel(self) -> str: ...
    def ConvertToRGB(self, translator: IColorTranslator) -> IAR_DataScatterPointsRgb: ...
    def FillPointValues(self, fullSize: int) -> tuple[list[float], list[float], list[float]]: ...
    def GetPoint(self, idx: int) -> IAR_ScatterPoint: ...

class IAR_DataScatterPointsRgb:
    @property
    def Description(self) -> str: ...
    @property
    def NumPoints(self) -> int: ...
    @property
    def Points(self) -> list[IAR_ScatterPointRgb]: ...
    @property
    def XLabel(self) -> str: ...
    @property
    def YLabel(self) -> str: ...
    @property
    def ValueLabel(self) -> str: ...
    def FillPointValues(
        self, fullSize: int
    ) -> tuple[list[float], list[float], list[float], list[float], list[float]]: ...
    def GetPoint(self, idx: int) -> IAR_ScatterPointRgb: ...

class IAR_DataSeries:
    @property
    def Description(self) -> str: ...
    @property
    def XLabel(self) -> str: ...
    @property
    def XData(self) -> IVectorData: ...
    @property
    def SeriesLabels(self) -> list[str]: ...
    @property
    def NumSeries(self) -> int: ...
    @property
    def YData(self) -> IMatrixData: ...
    def ConvertToRGB(self, translator: IColorTranslator) -> IAR_DataSeriesRgb: ...

class IAR_DataSeriesRgb:
    @property
    def Description(self) -> str: ...
    @property
    def NumSeries(self) -> int: ...
    @property
    def XData(self) -> IVectorData: ...
    @property
    def NumberOfRows(self) -> int: ...
    @property
    def YVals(self) -> list[list[IAR_Rgb]]: ...
    @property
    def XLabel(self) -> str: ...
    @property
    def SeriesLabels(self) -> list[str]: ...
    def FillYValues(self, fullSize: int) -> tuple[list[float], list[float], list[float]]: ...
    def GetYPoint(self, row: int, col: int) -> IAR_Rgb: ...

class IAR_HeaderData:
    @property
    def Lines(self) -> list[str]: ...

class IAR_MetaData:
    @property
    def FeatureDescription(self) -> str: ...
    @property
    def LensFile(self) -> str: ...
    @property
    def LensTitle(self) -> str: ...
    @property
    def Date(self) -> DateTime: ...

class IAR_NSCSingleRayTraceData:
    @property
    def IsValid(self) -> bool: ...
    @property
    def ZRDFile(self) -> str: ...
    @property
    def NumberOfSegments(self) -> int: ...
    @property
    def WaveIndex(self) -> int: ...
    @property
    def WavelengthUM(self) -> float: ...
    def ReadSegmentFull(
        self, segmentNumber: int
    ) -> tuple[
        bool,
        int,
        int,
        int,
        int,
        int,
        RayStatus,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        int,
        int,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
    ]: ...

class IAR_PathAnalysisData:
    @property
    def TotalFluxIn(self) -> float: ...
    @property
    def TotalFluxOut(self) -> float: ...
    @property
    def TotalRays(self) -> int: ...
    @property
    def TotalHits(self) -> int: ...
    @property
    def NumPaths(self) -> int: ...
    @property
    def PathNumber(self) -> int: ...
    @PathNumber.setter
    def PathNumber(self, value: int) -> None: ...
    @property
    def PathBranches(self) -> int: ...
    @property
    def PathSource(self) -> int: ...
    @property
    def LastObject(self) -> int: ...
    @property
    def LastFace(self) -> int: ...
    @property
    def PathFluxOut(self) -> float: ...
    @property
    def PathFluxPercent(self) -> float: ...
    @property
    def PathSequence(self) -> str: ...
    @property
    def Paths(self) -> list[IAR_PathAnalysisEntry]: ...
    def GetPathNumber(self, idx: int) -> IAR_PathAnalysisEntry: ...

class IAR_PathAnalysisEntry:
    @property
    def PathNumber(self) -> int: ...
    @property
    def RaysInPath(self) -> int: ...
    @property
    def HitsInPath(self) -> int: ...
    @property
    def GhostsInPath(self) -> int: ...
    @property
    def UniqueObjectsInPath(self) -> int: ...
    @property
    def TotalPathFlux(self) -> float: ...
    @property
    def NumberOfObjectsInPath(self) -> int: ...
    @property
    def PathObjectList(self) -> list[int]: ...
    @property
    def PathObjectAndFaceList(self) -> list[list[int]]: ...
    @property
    def PathSource(self) -> int: ...
    def GetPathObjectNumber(self, objectNumber: int) -> int: ...

class IAR_RayData:
    @property
    def Description(self) -> str: ...
    @property
    def NumRays(self) -> int: ...
    @property
    def Rays(self) -> list[IAR_RayInfo]: ...
    def GetRay(self, idx: int) -> IAR_RayInfo: ...

class IAR_RayInfo:
    @property
    def RayIndex(self) -> int: ...
    @property
    def Segment(self) -> int: ...
    @property
    def Parent(self) -> int: ...
    @property
    def Level(self) -> int: ...
    @property
    def In_object(self) -> int: ...
    @property
    def Hit_object(self) -> int: ...
    @property
    def Hit_face(self) -> int: ...
    @property
    def X(self) -> float: ...
    @property
    def Y(self) -> float: ...
    @property
    def Z(self) -> float: ...
    @property
    def L(self) -> float: ...
    @property
    def M(self) -> float: ...
    @property
    def N(self) -> float: ...
    @property
    def Nx(self) -> float: ...
    @property
    def Ny(self) -> float: ...
    @property
    def Nz(self) -> float: ...
    @property
    def Ex(self) -> float: ...
    @property
    def Ey(self) -> float: ...
    @property
    def Ez(self) -> float: ...
    @property
    def PathLength(self) -> float: ...
    @property
    def OpticalPathLength(self) -> float: ...
    @property
    def Wavelength(self) -> float: ...
    @property
    def Vignetted(self) -> int: ...
    @property
    def Error(self) -> int: ...

class IAR_Rgb:
    @property
    def R(self) -> float: ...
    @property
    def G(self) -> float: ...
    @property
    def B(self) -> float: ...

class IAR_ScatterPoint:
    @property
    def X(self) -> float: ...
    @property
    def Y(self) -> float: ...
    @property
    def Value(self) -> float: ...

class IAR_ScatterPointRgb:
    @property
    def X(self) -> float: ...
    @property
    def Y(self) -> float: ...
    @property
    def Value(self) -> IAR_Rgb: ...

class IAR_SpotDataResult:
    @property
    def X(self) -> float: ...
    @X.setter
    def X(self, value: float) -> None: ...
    @property
    def Y(self) -> float: ...
    @Y.setter
    def Y(self, value: float) -> None: ...
    @property
    def Z(self) -> float: ...
    @Z.setter
    def Z(self, value: float) -> None: ...
    @property
    def L(self) -> float: ...
    @L.setter
    def L(self, value: float) -> None: ...
    @property
    def M(self) -> float: ...
    @M.setter
    def M(self, value: float) -> None: ...
    @property
    def N(self) -> float: ...
    @N.setter
    def N(self, value: float) -> None: ...
    @property
    def Detector_X(self) -> float: ...
    @Detector_X.setter
    def Detector_X(self, value: float) -> None: ...
    @property
    def Detector_Y(self) -> float: ...
    @Detector_Y.setter
    def Detector_Y(self, value: float) -> None: ...
    @property
    def Detector_Z(self) -> float: ...
    @Detector_Z.setter
    def Detector_Z(self, value: float) -> None: ...
    @property
    def RefCoord_X(self) -> float: ...
    @RefCoord_X.setter
    def RefCoord_X(self, value: float) -> None: ...
    @property
    def RefCoord_Y(self) -> float: ...
    @RefCoord_Y.setter
    def RefCoord_Y(self, value: float) -> None: ...
    @property
    def RMSSpot_X(self) -> float: ...
    @RMSSpot_X.setter
    def RMSSpot_X(self, value: float) -> None: ...
    @property
    def RMSSpot_Y(self) -> float: ...
    @RMSSpot_Y.setter
    def RMSSpot_Y(self, value: float) -> None: ...
    @property
    def RMSSpotSize(self) -> float: ...
    @RMSSpotSize.setter
    def RMSSpotSize(self, value: float) -> None: ...
    @property
    def GeoSpotSize(self) -> float: ...
    @GeoSpotSize.setter
    def GeoSpotSize(self, value: float) -> None: ...

class IAR_SpotDataResultMatrix:
    @property
    def HalfWidth_X(self) -> float: ...
    @property
    def HalfWidth_Y(self) -> float: ...
    @property
    def MaxRadius(self) -> float: ...
    @property
    def MeanRadius(self) -> float: ...
    @property
    def NumberOfFields(self) -> int: ...
    @property
    def NumberOfWavelengths(self) -> int: ...
    def Get_L_For(self, fieldN: int, waveN: int) -> float: ...
    def Get_M_For(self, fieldN: int, waveN: int) -> float: ...
    def Get_N_For(self, fieldN: int, waveN: int) -> float: ...
    def Get_X_For(self, fieldN: int, waveN: int) -> float: ...
    def Get_Y_For(self, fieldN: int, waveN: int) -> float: ...
    def Get_Z_For(self, fieldN: int, waveN: int) -> float: ...
    def GetDetector_X_For(self, fieldN: int, waveN: int) -> float: ...
    def GetDetector_Y_For(self, fieldN: int, waveN: int) -> float: ...
    def GetDetector_Z_For(self, fieldN: int, waveN: int) -> float: ...
    def GetGeoSpotSizeFor(self, fieldN: int, waveN: int) -> float: ...
    def GetReferenceCoordinate_X_For(self, fieldN: int, waveN: int) -> float: ...
    def GetReferenceCoordinate_Y_For(self, fieldN: int, waveN: int) -> float: ...
    def GetRMSSpot_X_For(self, fieldN: int, waveN: int) -> float: ...
    def GetRMSSpot_Y_For(self, fieldN: int, waveN: int) -> float: ...
    def GetRMSSpotSizeFor(self, fieldN: int, waveN: int) -> float: ...

class IAR_Vector2D:
    @property
    def dx(self) -> float: ...
    @property
    def dy(self) -> float: ...

class IAR_XYZ:
    @property
    def X(self) -> float: ...
    @property
    def Y(self) -> float: ...
    @property
    def Z(self) -> float: ...
